# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
import typing
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

import aiosmtplib

import mlrun.common.schemas
import mlrun.lists
import mlrun.utils.helpers
import mlrun.utils.notifications.notification.base as base
import mlrun.utils.regex

DEFAULT_SMTP_PORT = 587


class MailNotification(base.NotificationBase):
    """
    API/Client notification for sending run statuses as a mail message
    """

    boolean_params = ["use_tls", "start_tls", "validate_certs"]

    required_params = [
        "server_host",
        "server_port",
        "sender_address",
        "username",
        "password",
        "email_addresses",
    ] + boolean_params

    @classmethod
    def validate_params(cls, params):
        for required_param in cls.required_params:
            if required_param not in params:
                raise ValueError(
                    f"Parameter '{required_param}' is required for MailNotification"
                )

        for boolean_param in cls.boolean_params:
            if not isinstance(params.get(boolean_param, None), bool):
                raise ValueError(
                    f"Parameter '{boolean_param}' must be a boolean for MailNotification"
                )

        cls._validate_emails(params)

    async def push(
        self,
        message: str,
        severity: typing.Optional[
            typing.Union[mlrun.common.schemas.NotificationSeverity, str]
        ] = mlrun.common.schemas.NotificationSeverity.INFO,
        runs: typing.Optional[typing.Union[mlrun.lists.RunList, list]] = None,
        custom_html: typing.Optional[typing.Optional[str]] = None,
        alert: typing.Optional[mlrun.common.schemas.AlertConfig] = None,
        event_data: typing.Optional[mlrun.common.schemas.Event] = None,
    ):
        self.params["subject"] = f"[{severity}] {message}"
        message_body_override = self.params.get("message_body_override", None)

        runs_html = self._get_html(
            message, severity, runs, custom_html, alert, event_data
        )
        self.params["body"] = runs_html

        if message_body_override:
            self.params["body"] = message_body_override.replace(
                "{{ runs }}", runs_html
            ).replace("{{runs}}", runs_html)

        await self._send_email(**self.params)

    @classmethod
    def enrich_default_params(
        cls, params: dict, default_params: typing.Optional[dict] = None
    ) -> dict:
        params = super().enrich_default_params(params, default_params)
        params.setdefault("use_tls", True)
        params.setdefault("start_tls", False)
        params.setdefault("validate_certs", True)
        params.setdefault("server_port", DEFAULT_SMTP_PORT)

        default_mail_address = params.pop("default_email_addresses", "")
        params["email_addresses"] = cls._merge_mail_addresses(
            default_mail_address, params.get("email_addresses", "")
        )

        return params

    @classmethod
    def _merge_mail_addresses(
        cls,
        default_mail_address: typing.Union[str, list],
        email_addresses: typing.Union[str, list],
    ) -> str:
        if isinstance(default_mail_address, str):
            default_mail_address = (
                default_mail_address.split(",") if default_mail_address else []
            )
        if isinstance(email_addresses, str):
            email_addresses = email_addresses.split(",") if email_addresses else []
        email_addresses.extend(default_mail_address)
        email_addresses_str = ",".join(email_addresses)
        return email_addresses_str

    @classmethod
    def _validate_emails(cls, params):
        cls._validate_email_address(params["sender_address"])

        if not isinstance(params["email_addresses"], (str, list)):
            raise ValueError(
                "Parameter 'email_addresses' must be a string or a list of strings"
            )

        email_addresses = params["email_addresses"]
        if isinstance(email_addresses, str):
            email_addresses = email_addresses.split(",")
        for email_address in email_addresses:
            cls._validate_email_address(email_address)

    @classmethod
    def _validate_email_address(cls, email_address):
        if not isinstance(email_address, str):
            raise ValueError(f"Email address '{email_address}' must be a string")

        if not re.match(mlrun.utils.regex.mail_regex, email_address):
            raise ValueError(f"Invalid email address '{email_address}'")

    @staticmethod
    async def _send_email(
        email_addresses: str,
        sender_address: str,
        server_host: str,
        server_port: int,
        username: str,
        password: str,
        use_tls: bool,
        start_tls: bool,
        validate_certs: bool,
        subject: str,
        body: str,
        **kwargs,
    ):
        # Create the email message
        message = MIMEMultipart("alternative")
        message["From"] = sender_address
        message["To"] = email_addresses
        message["Subject"] = subject
        message.attach(MIMEText(body, "html"))

        # Send the email
        await aiosmtplib.send(
            message,
            hostname=server_host,
            port=server_port,
            username=username,
            password=password,
            use_tls=use_tls,
            validate_certs=validate_certs,
            start_tls=start_tls,
        )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import sys
from typing import Callable, Union

import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots

import mlrun.common.schemas.model_monitoring
from mlrun.artifacts import PlotlyArtifact

# A type for representing a drift result, a tuple of the status and the drift mean:
DriftResultType = tuple[
    mlrun.common.schemas.model_monitoring.constants.ResultStatusApp, float
]


class _PlotlyTableArtifact(PlotlyArtifact):
    """A custom class for plotly table artifacts"""

    @staticmethod
    def _disable_table_dragging(figure_html: str) -> str:
        """
        Disable the table columns dragging by adding the following
        JavaScript code
        """
        start, end = figure_html.rsplit(";", 1)
        middle = (
            ';for (const element of document.getElementsByClassName("table")) '
            '{element.style.pointerEvents = "none";}'
        )
        figure_html = start + middle + end
        return figure_html

    def get_body(self) -> str:
        """Get the adjusted HTML representation of the figure"""
        return self._disable_table_dragging(super().get_body())


class FeaturesDriftTablePlot:
    """
    Class for producing a features drift table. The plot is a table with columns of all the statistics and metrics
    provided with two additional plot columns of the histograms and drift status. The rows content will be drawn
    per feature.

    For example, if the statistics are 'mean', 'min', 'max' and one metric of 'tvd', for 3 features the table will be:

    | feature name |       mean     |       min      |       max      |  tvd  |   histograms   |   |
    |______________|________________|________________|________________|_______|________________|___|
    |              | sample | input | sample | input | sample | input |       |                |   |
    |______________|________|_______|________|_______|________|_______|_______|________________|___|
    | feature 1    | 0.1    | 0.15  | 0.02   | 0.03  | 12.0   | 15.8  | 0.1   | ...            | V |
    | feature 2    | 0.1    | 0.15  | 0.02   | 0.03  | 12.0   | 15.8  | 0.1   | ...            | V |
    | feature 3    | 0.1    | 0.15  | 0.02   | 0.03  | 12.0   | 15.8  | 0.1   | ...            | V |
    """

    # Table column widths:
    _FEATURE_NAME_COLUMN_WIDTH = 140
    _VALUE_COLUMN_WIDTH = (
        70  # The width for the values of all the statistics and metrics columns.
    )
    _HISTOGRAMS_COLUMN_WIDTH = 180
    _STATUS_COLUMN_WIDTH = 20

    # Table rows heights:
    _HEADER_ROW_HEIGHT = 25
    _FEATURE_ROW_HEIGHT = 50  # Will be increased depends on the longest feature name.

    # Histograms configurations:
    _SAMPLE_SET_HISTOGRAM_COLOR = "rgb(0,112,192)"  # Blue
    _INPUTS_HISTOGRAM_COLOR = "rgb(208,0,106)"  # Magenta
    _HISTOGRAM_OPACITY = 0.75

    # Status configurations:
    _STATUS_COLORS = {
        mlrun.common.schemas.model_monitoring.constants.ResultStatusApp.no_detection: "rgb(0,176,80)",  # Green
        mlrun.common.schemas.model_monitoring.constants.ResultStatusApp.potential_detection: "rgb(255,192,0)",  # Orange
        mlrun.common.schemas.model_monitoring.constants.ResultStatusApp.detected: "rgb(208,0,106)",  # Magenta
    }

    # Font configurations:
    _FONT_SIZE = 10
    _FONT_COLOR = "rgb(68,68,68)"  # Dark Grey

    # General configurations:
    _FEATURE_NAME_MAX_LENGTH = (
        16  # Each 16 characters the feature name will continue in a new line.
    )
    _VALUES_MAX_DIGITS = (
        6  # Each of the statistics and metrics will be rounded to show 6 digits.
    )
    _BACKGROUND_COLOR = "rgb(255,255,255)"  # White
    _SEPARATORS_COLOR = "rgb(240,240,240)"  # Light grey

    def __init__(self):
        """
        Initialize the plot producer for later calling the `produce` method.
        """
        # Initialize columns names lists for the table that must be read before plotting:
        self._statistics_columns = None  # type: List[str]
        self._metrics_columns = None  # type: List[str]
        self._plot_columns = None  # type: List[str]
        self._value_columns_widths = None  # type: List[int]

    def produce(
        self,
        sample_set_statistics: dict,
        inputs_statistics: dict,
        metrics: dict[str, Union[dict, float]],
        drift_results: dict[str, DriftResultType],
    ) -> _PlotlyTableArtifact:
        """
        Produce the html code of the table plot with the given information and the stored configurations in the class.

        :param sample_set_statistics: The sample set calculated statistics dictionary.
        :param inputs_statistics:     The inputs calculated statistics dictionary.
        :param metrics:               The drift detection metrics calculated on the sample set and inputs.
        :param drift_results:         The drift results per feature according to the rules of the monitor.

        :return: The drift table as a plotly artifact.
        """
        figure = self._plot(
            features=list(inputs_statistics.keys()),
            sample_set_statistics=sample_set_statistics,
            inputs_statistics=inputs_statistics,
            metrics=metrics,
            drift_results=drift_results,
        )
        return _PlotlyTableArtifact(figure=figure, key="drift_table_plot")

    def _read_columns_names(self, statistics_dictionary: dict, drift_metrics: dict):
        """
        Read the available statistics and metrics to include them as columns in the table.

        :param statistics_dictionary: A statistics dictionary (one of sample_set or inputs).
        :param drift_metrics:         The metrics dictionary.
        """
        # Take all statistics from the first feature:
        self._statistics_columns = list(list(statistics_dictionary.values())[0].keys())
        self._statistics_columns.remove("hist")

        # Take all the metrics from the first feature:
        self._metrics_columns = list(list(drift_metrics.values())[0].keys())

        # Add the 'histograms' column if available:
        self._plot_columns = ["histograms", "detection"]

    def _calculate_columns_widths(self):
        """
        Calculate the column widths to draw the table. Can be called only after the method `_read_columns_names` was
        called.
        """
        # For the statistics, 2x`value_width` as there will be two values, one for the sample and one for the input:
        self._value_columns_widths = [2 * self._VALUE_COLUMN_WIDTH] * len(
            self._statistics_columns
        )

        # For the metrics:
        self._value_columns_widths += [self._VALUE_COLUMN_WIDTH] * len(
            self._metrics_columns
        )

    def _plot_headers_tables(self) -> tuple[go.Table, go.Table]:
        """
        Plot the headers of the table:

        * The main header with the column names.
        * The sub-header with the statistics column splits to 'sample | inputs'.

        :return: A tuple with both headers - `Table` traces:
                 [0] - Header.
                 [1] - Sub-header.
        """
        # Generate the header of the table:
        headers = [
            # [0] - The feature name column
            "",
            # [1:-2] - Statistics and metrics columns.
            *self._statistics_columns,
            *self._metrics_columns,
            # [-2] - The histograms column
            "histograms",
            # [-1] - The notifications column
            "",
        ]
        header_table = go.Table(
            header={
                "values": [
                    f"<b>{header.capitalize()}</b>" for header in headers
                ],  # Make the text bold and with starting capital letter.
                "align": "center",
                "font": {"size": self._FONT_SIZE, "color": self._FONT_COLOR},
                "line": {"color": self._SEPARATORS_COLOR},
            },
            columnwidth=[
                self._FEATURE_NAME_COLUMN_WIDTH,
                *self._value_columns_widths,
                self._HISTOGRAMS_COLUMN_WIDTH,
                self._STATUS_COLUMN_WIDTH,
            ],
            header_fill_color=self._BACKGROUND_COLOR,
        )

        # Generate the sub-headers (for each of the statistics column there should be two columns, one for the sample
        # set and one for the inputs):
        sub_headers = (
            [""]
            + ["Sample", "Input"] * len(self._statistics_columns)
            + [""] * len(self._metrics_columns)
            + ["", ""]
        )
        sub_header_table = go.Table(
            header={
                "values": sub_headers,
                "align": "center",
                "font": {"size": self._FONT_SIZE, "color": self._FONT_COLOR},
                "line": {"color": self._SEPARATORS_COLOR},
            },
            columnwidth=(
                [self._FEATURE_NAME_COLUMN_WIDTH]
                + [self._VALUE_COLUMN_WIDTH]
                * (2 * len(self._statistics_columns) + len(self._metrics_columns))
                + [self._HISTOGRAMS_COLUMN_WIDTH, self._STATUS_COLUMN_WIDTH]
            ),
            header_fill_color=self._BACKGROUND_COLOR,
        )

        return header_table, sub_header_table

    def _separate_feature_name(self, feature_name: str) -> list[str]:
        """
        Separate the given feature name by the maximum length configured in the class. Used for calculating the amount
        of lines required to represent the longest feature name in the table, so the row heights will fit accordingly.

        :param feature_name: The feature name to separate.

        :return: The feature name's list of the separations.
        """
        return [
            feature_name[i : i + self._FEATURE_NAME_MAX_LENGTH]
            for i in range(0, len(feature_name), self._FEATURE_NAME_MAX_LENGTH)
        ]

    def _get_value_format(self, value: Union[str, int, float]) -> str:
        """
        Plotly uses D3 formatter to format values. This method return the format according to the configured
        properties in the class to the given value - one of the cells in a feature row statistics and metrics values.

        :param value: A value of a feature row cell.

        :return: The value's format.
        """
        # A string does not need reformatting:
        if isinstance(value, str):
            return ""

        # Check fo nan values:
        if np.isnan(value):
            return ""

        # Any whole number or number with a long integer value should be parsed into short characters
        # (e.g: 10000 -> 10k, 1100000 -> 1.1m):
        integer_length = len(str(int(value)))
        if value % 1 == 0 or integer_length > self._VALUES_MAX_DIGITS:
            return "~s"

        # Round it to the remaining digits left considering the integer length:
        return f".{self._VALUES_MAX_DIGITS - integer_length}"

    def _plot_feature_row_table(
        self,
        feature_name: str,
        sample_statistics: dict,
        input_statistics: dict,
        metrics: dict,
        row_height: int,
    ) -> go.Table:
        """
        Plot the feature row to include in the table. The row will include only the columns of the statistics and
        metrics values. The histogram and drift notification are plotted in different methods.

        :param feature_name:      The feature's name.
        :param sample_statistics: The feature's sample set statistics dictionary.
        :param input_statistics:  The feature's inputs statistics dictionary.
        :param metrics:           The feature's metrics results dictionary.
        :param row_height:        The height of the row to plot.

        :return: The feature row - `Table` trace.
        """
        # Add '\n' to the feature name in order to make it fit into its cell:
        html_feature_name = "<br>".join(self._separate_feature_name(feature_name))

        # Initialize the cells values list with the bold feature name as the first value:
        cells_values = [f"<b>{html_feature_name}</b>"]

        # Add the statistics columns:
        for column in self._statistics_columns:
            cells_values.append(sample_statistics[column])
            try:
                cells_values.append(input_statistics[column])
            except KeyError:
                raise ValueError(
                    f"The `input_statistics['{feature_name}']` dictionary "
                    f"does not include the expected key '{column}'. "
                    "Please check the current data."
                )

        # Add the metrics columns:
        for column in self._metrics_columns:
            cells_values.append(metrics[column])

        # Get the cells values formats:
        cells_formats = [
            self._get_value_format(value=cell_value) for cell_value in cells_values
        ]

        # Create the row:
        feature_row_table = go.Table(
            header={
                "values": cells_values,
                "align": "center",
                "font": {"size": self._FONT_SIZE, "color": self._FONT_COLOR},
                "line": {"color": self._SEPARATORS_COLOR},
                "height": row_height,
                "format": cells_formats,
            },
            columnwidth=[self._FEATURE_NAME_COLUMN_WIDTH, self._VALUE_COLUMN_WIDTH],
            header_fill_color=self._BACKGROUND_COLOR,
        )

        return feature_row_table

    def _plot_histogram_bars(
        self,
        figure_add_trace: Callable,
        sample_hist: tuple[list, list],
        input_hist: tuple[list, list],
        showlegend: bool = False,
    ) -> None:
        """
        Plot the feature's histograms to include in the "histograms" column. Both histograms are rescaled
        to be from 0.0 to 1.0, so they will be drawn in the same scale regardless the amount of elements they were
        calculated upon.

        :param figure_add_trace: The figure's method that get the histogram and adds it to the figure.
        :param sample_hist:      The sample set histogram data.
        :param input_hist:       The input histogram data.
        :param showlegend:       Show the legend for each histogram or not.

        :return: None
        """
        # Plot the histograms:
        for name, color, histogram in zip(
            ["sample", "input"],
            [self._SAMPLE_SET_HISTOGRAM_COLOR, self._INPUTS_HISTOGRAM_COLOR],
            [sample_hist, input_hist],
        ):
            # Read the histogram tuple:
            counts, bins = histogram
            # Rescale the counts to be in percentages (between 0.0 to 1.0):
            counts = np.array(counts) / sum(counts)
            hovertext = [""] * len(counts)
            # Convert to NumPy for vectorization:
            bins = np.array(bins)
            if bins[0] == -sys.float_info.max:
                bins[0] = bins[1] - (bins[2] - bins[1])
                hovertext[0] = f"(-inf, {bins[1]})"
            if bins[-1] == sys.float_info.max:
                bins[-1] = bins[-2] + (bins[-2] - bins[-3])
                hovertext[-1] = f"({bins[-2]}, inf)"
            # Center the bins (leave the first one):
            bins = 0.5 * (bins[:-1] + bins[1:])
            # Plot the histogram as a line with filled background below it:
            histogram_bar = go.Bar(
                x=bins,
                y=counts,
                name=name,
                marker_color=color,
                opacity=self._HISTOGRAM_OPACITY,
                legendgroup=name,
                hovertext=hovertext,
                showlegend=showlegend,
            )
            figure_add_trace(histogram_bar)

    def _calculate_row_height(self, features: list[str]) -> int:
        """
        Calculate the feature row height according to the given features. The longest feature will set the height to all
        the rows. The height depends on the separations amount of the longest feature name - more '\n' means more pixels
        for the row's height to show the entirety of the feature name.

        :param features: The list of features.

        :return: The row height.
        """
        feature_name_seperations = max(
            [
                len(self._separate_feature_name(feature_name))
                for feature_name in features
            ]
        )
        return max(
            self._FEATURE_ROW_HEIGHT, 1.5 * self._FONT_SIZE * feature_name_seperations
        )

    def _plot_status_circle(
        self,
        figure: go.Figure,
        row: int,
        row_height: int,
        drift_result: DriftResultType,
    ):
        """
        Plot the drift status - a little circle with color as configured in the
        class. The color will be chosen according to the drift status given.

        :param figure:       The figure (feature row cell) to draw the circle in.
        :param row:          The row number.
        :param row_height:   The row height.
        :param drift_result: The drift result.
        """
        # Update the x-axis and y-axis to be in range of the `row_height` x `self._NOTIFICATIONS_COLUMN_WIDTH` so the
        # circle will be sized correctly. The axis number is equal to the row number - 3 (plots with axes starts at
        # row 3) times the plot columns (2 columns has axes in each row) + 2 (to get to the column of the notification):
        axis_number = (row - 3) * 2 + 2
        figure["layout"][f"xaxis{axis_number}"].update(
            range=[0, self._STATUS_COLUMN_WIDTH]
        )
        figure["layout"][f"yaxis{axis_number}"].update(range=[0, row_height])

        # Get the color:
        notification_color = self._STATUS_COLORS[drift_result[0]]
        half_transparent_notification_color = notification_color.replace(
            "rgb", "rgba"
        ).replace(")", ",0.5)")

        # Calculate the circle points adjusted to the given row height (aimed to where the text starts and to be at the
        # size of the text as well):
        y0 = 36 + (row_height - self._FEATURE_ROW_HEIGHT)
        y1 = y0 + self._FONT_SIZE
        x0 = (self._STATUS_COLUMN_WIDTH / 2) - ((y1 - y0) / 2)
        x1 = (self._STATUS_COLUMN_WIDTH / 2) + ((y1 - y0) / 2)

        # Draw the circle on top of the figure:
        figure.add_shape(
            type="circle",
            xref="x",
            yref="y",
            fillcolor=half_transparent_notification_color,
            x0=x0,
            y0=y0,
            x1=x1,
            y1=y1,
            line_color=notification_color,
            row=row,
            col=3,
        )

    def _plot(
        self,
        features: list[str],
        sample_set_statistics: dict,
        inputs_statistics: dict,
        metrics: dict[str, Union[dict, float]],
        drift_results: dict[str, DriftResultType],
    ) -> go.Figure:
        """
        Plot the drift table using the given data and stored configurations of the class.

        :param features:              List of all the features names to include in the table. These names expected to be
                                      in the statistics and metrics dictionaries.
        :param sample_set_statistics: The sample set calculated statistics dictionary.
        :param inputs_statistics:     The inputs calculated statistics dictionary.
        :param metrics:               The drift detection metrics calculated on the sample set and inputs.
        :param drift_results:         The drift results per feature according to the rules of the monitor.

        :return: The plot's figure.
        """
        # Read and calculate widths and heights for the table:
        self._read_columns_names(
            statistics_dictionary=sample_set_statistics, drift_metrics=metrics
        )
        self._calculate_columns_widths()
        row_height = self._calculate_row_height(features=features)
        rows = 2 + len(features)  # Header + sub-header + row per feature
        columns = 3  # Values table + histograms plot + detection plot

        # Calculate the height and width of the entire plot:
        width = (
            self._FEATURE_NAME_COLUMN_WIDTH
            + sum(self._value_columns_widths)
            + self._HISTOGRAMS_COLUMN_WIDTH
            + self._STATUS_COLUMN_WIDTH
        )
        height = 2 * self._HEADER_ROW_HEIGHT + len(features) * row_height

        # Create the main figure - a subplots figure with the calculated rows and columns (the `row_heights` and
        # `column_widths` attributes here are calculated in percentages):
        main_figure = make_subplots(
            rows=rows,
            cols=columns,
            specs=(
                [[{"type": "table", "colspan": 3}, None, None]] * 2
                + [[{"type": "table"}, {}, {}]] * (rows - 2)
            ),
            row_heights=(
                [self._HEADER_ROW_HEIGHT / height] * 2
                + [row_height / height] * (rows - 2)
            ),
            column_widths=[
                (self._FEATURE_NAME_COLUMN_WIDTH + sum(self._value_columns_widths))
                / width,
                self._HISTOGRAMS_COLUMN_WIDTH / width,
                self._STATUS_COLUMN_WIDTH / width,
            ],
            horizontal_spacing=0,
            vertical_spacing=0,
        )

        # Add the first two rows - the header and sub-header:
        header_trace, sub_header_trace = self._plot_headers_tables()
        main_figure.add_trace(header_trace, row=1, col=1)
        main_figure.add_trace(sub_header_trace, row=2, col=1)

        # Start going over the features and plot each row, histogram and status
        for row, feature in enumerate(
            features,
            start=3,  # starting from row 3 after the headers
        ):
            try:
                # Add the feature values:
                main_figure.add_trace(
                    self._plot_feature_row_table(
                        feature_name=feature,
                        sample_statistics=sample_set_statistics[feature],
                        input_statistics=inputs_statistics[feature],
                        metrics=metrics[feature],
                        row_height=row_height,
                    ),
                    row=row,
                    col=1,
                )
            except KeyError:
                raise ValueError(
                    "`sample_set_statistics` does not contain the expected "
                    f"key '{feature}' from `inputs_statistics`. Please verify "
                    "the data integrity.\n"
                    f"{sample_set_statistics.keys() = }\n"
                    f"{inputs_statistics.keys() = }\n"
                )
            # Add the histograms (both traces are added to the same subplot figure):
            self._plot_histogram_bars(
                figure_add_trace=functools.partial(
                    main_figure.add_trace, row=row, col=2
                ),
                sample_hist=sample_set_statistics[feature]["hist"],
                input_hist=inputs_statistics[feature]["hist"],
                # Only the first row should have its legend visible
                showlegend=(row == 3),
            )
            # Add the status (a circle with color according to the drift status)
            self._plot_status_circle(
                figure=main_figure,
                row=row,
                row_height=row_height,
                drift_result=drift_results[feature],
            )

        # Configure the layout and axes for height and widths:
        main_figure.update_layout(
            autosize=False,
            width=width,
            height=height,
            plot_bgcolor=self._BACKGROUND_COLOR,
            paper_bgcolor=self._BACKGROUND_COLOR,
            margin={"t": 0, "b": 0, "l": 0, "r": 0, "autoexpand": False},
            font_size=self._FONT_SIZE,  # Control the font of the x and y axes and the legend.
            font_color=self._FONT_COLOR,
            legend={
                "orientation": "h",
                "yanchor": "top",
                "y": 1.0 - (self._HEADER_ROW_HEIGHT / height) + 0.002,
                "xanchor": "right",
                "x": 1.0 - (self._STATUS_COLUMN_WIDTH / width) - 0.01,
                "bgcolor": "rgba(0,0,0,0)",
            },
            barmode="overlay",
            bargap=0,
        )
        main_figure.update_xaxes(
            showticklabels=False,
            showline=True,
            linewidth=1,
            linecolor=self._SEPARATORS_COLOR,
            mirror=True,
        )
        main_figure.update_yaxes(
            showticklabels=False,
            showline=True,
            linewidth=1,
            linecolor=self._SEPARATORS_COLOR,
            mirror=True,
        )

        return main_figure



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .constants import (
    INTERSECT_DICT_KEYS,
    ApplicationEvent,
    DriftStatus,
    EndpointMode,
    EndpointType,
    EndpointUID,
    EventFieldType,
    EventKeyMetrics,
    EventLiveStats,
    FeatureSetFeatures,
    FileTargetKind,
    FunctionURI,
    MetricData,
    ModelEndpointCreationStrategy,
    ModelEndpointMonitoringMetricType,
    ModelEndpointSchema,
    ModelMonitoringAppLabel,
    ModelMonitoringInfraLabel,
    ModelMonitoringMode,
    MonitoringFunctionNames,
    PredictionsQueryConstants,
    ProjectSecretKeys,
    ResultData,
    ResultKindApp,
    ResultStatusApp,
    SpecialApps,
    StreamProcessingEvent,
    TDEngineSuperTables,
    TSDBTarget,
    V3IOTSDBTables,
    VersionedModel,
    WriterEvent,
    WriterEventKind,
)
from .functions import FunctionsType, FunctionSummary
from .grafana import (
    GrafanaColumn,
    GrafanaColumnType,
    GrafanaNumberColumn,
    GrafanaStringColumn,
    GrafanaTable,
)
from .model_endpoints import (
    ApplicationMetricRecord,
    ApplicationResultRecord,
    Features,
    FeatureValues,
    ModelEndpoint,
    ModelEndpointDriftValues,
    ModelEndpointList,
    ModelEndpointMetadata,
    ModelEndpointMonitoringMetric,
    ModelEndpointMonitoringMetricNoData,
    ModelEndpointMonitoringMetricValues,
    ModelEndpointMonitoringResultValues,
    ModelEndpointSpec,
    ModelEndpointStatus,
)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import inspect
import os
import re
import time
import typing
import warnings
from collections.abc import Iterable
from enum import Enum
from typing import Optional

import dotenv
import kubernetes.client as k8s_client
from kubernetes.client import V1Volume, V1VolumeMount

import mlrun.common.constants
import mlrun.common.secrets
import mlrun.errors
import mlrun.runtimes.mounts
import mlrun.utils.regex
from mlrun.common.schemas import (
    PreemptionModes,
    SecurityContextEnrichmentModes,
)

from ..config import config as mlconf
from ..k8s_utils import (
    generate_preemptible_nodes_affinity_terms,
    validate_node_selectors,
)
from ..utils import logger, update_in
from .base import BaseRuntime, FunctionSpec, spec_fields
from .utils import (
    get_gpu_from_resource_requirement,
    get_item_name,
    set_named_item,
    verify_limits,
    verify_requests,
)

# TODO: add env attribute to the sanitized types
sanitized_types = {
    "affinity": {
        "attribute_type_name": "V1Affinity",
        "attribute_type": k8s_client.V1Affinity,
        "sub_attribute_type": None,
        "contains_many": False,
        "not_sanitized_class": dict,
    },
    "tolerations": {
        "attribute_type_name": "List[V1.Toleration]",
        "attribute_type": list,
        "contains_many": True,
        "sub_attribute_type": k8s_client.V1Toleration,
        "not_sanitized_class": list,
    },
    "security_context": {
        "attribute_type_name": "V1SecurityContext",
        "attribute_type": k8s_client.V1SecurityContext,
        "sub_attribute_type": None,
        "contains_many": False,
        "not_sanitized_class": dict,
    },
}

sanitized_attributes = {
    "affinity": sanitized_types["affinity"],
    "tolerations": sanitized_types["tolerations"],
    "security_context": sanitized_types["security_context"],
    "executor_tolerations": sanitized_types["tolerations"],
    "driver_tolerations": sanitized_types["tolerations"],
    "executor_affinity": sanitized_types["affinity"],
    "driver_affinity": sanitized_types["affinity"],
    "executor_security_context": sanitized_types["security_context"],
    "driver_security_context": sanitized_types["security_context"],
}


class KubeResourceSpec(FunctionSpec):
    _dict_fields = spec_fields + [
        "volumes",
        "volume_mounts",
        "env",
        "resources",
        "replicas",
        "image_pull_policy",
        "service_account",
        "image_pull_secret",
        "node_name",
        "node_selector",
        "affinity",
        "priority_class_name",
        "tolerations",
        "preemption_mode",
        "security_context",
        "state_thresholds",
        "serving_spec",
        "track_models",
        "parameters",
        "graph",
        "filename",
    ]
    _default_fields_to_strip = FunctionSpec._default_fields_to_strip + [
        "volumes",
        "volume_mounts",
        "resources",
        "replicas",
        "image_pull_policy",
        "service_account",
        "image_pull_secret",
        "node_name",
        "node_selector",
        "affinity",
        "priority_class_name",
        "tolerations",
        "preemption_mode",
        "security_context",
    ]
    _k8s_fields_to_serialize = [
        "volumes",
        "volume_mounts",
        "resources",
        "env",
        "image_pull_policy",
        "service_account",
        "image_pull_secret",
        "node_name",
        "node_selector",
        "affinity",
        "tolerations",
        "security_context",
    ]
    _fields_to_serialize = FunctionSpec._fields_to_serialize + _k8s_fields_to_serialize
    _fields_to_enrich = FunctionSpec._fields_to_enrich + [
        "env",  # Removing sensitive data from env
    ]
    _fields_to_skip_validation = FunctionSpec._fields_to_skip_validation + [
        # TODO: affinity, tolerations and node_selector are skipped due to preemption mode transitions.
        #  Preemption mode 'none' depends on the previous mode while the default mode may enrich these values.
        #  When we allow 'None' values for these attributes we get their true values and they will undo the default
        #  enrichment when creating the runtime from dict.
        #  The enrichment should move to the server side and then this can be removed.
        "affinity",
        "tolerations",
        "node_selector",
    ]

    def __init__(
        self,
        command=None,
        args=None,
        image=None,
        mode=None,
        volumes=None,
        volume_mounts=None,
        env=None,
        resources=None,
        default_handler=None,
        pythonpath=None,
        entry_points=None,
        description=None,
        workdir=None,
        replicas=None,
        image_pull_policy=None,
        service_account=None,
        build=None,
        image_pull_secret=None,
        node_name=None,
        node_selector=None,
        affinity=None,
        disable_auto_mount=False,
        priority_class_name=None,
        tolerations=None,
        preemption_mode=None,
        security_context=None,
        state_thresholds=None,
        serving_spec=None,
        track_models=None,
        parameters=None,
        graph=None,
    ):
        super().__init__(
            command=command,
            args=args,
            image=image,
            mode=mode,
            build=build,
            entry_points=entry_points,
            description=description,
            workdir=workdir,
            default_handler=default_handler,
            pythonpath=pythonpath,
            disable_auto_mount=disable_auto_mount,
        )
        self._volumes = {}
        self._volume_mounts = {}
        self.volumes = volumes or []
        self.volume_mounts = volume_mounts or []
        # TODO: add env attribute to the sanitized types
        self.env = env or []
        self._resources = self.enrich_resources_with_default_pod_resources(
            "resources", resources
        )

        self.replicas = replicas
        self.image_pull_policy = image_pull_policy
        # default service account is set in mlrun.utils.process_function_service_account
        # due to project specific defaults
        self.service_account = service_account
        self.image_pull_secret = image_pull_secret
        self.node_name = node_name
        self.node_selector = node_selector or {}
        self._affinity = affinity
        self.priority_class_name = (
            priority_class_name or mlrun.mlconf.default_function_priority_class_name
        )
        self._tolerations = tolerations
        self.preemption_mode = preemption_mode
        self.security_context = (
            security_context or mlrun.mlconf.get_default_function_security_context()
        )
        self.state_thresholds = (
            state_thresholds
            or mlrun.mlconf.function.spec.state_thresholds.default.to_dict()
        )
        self.serving_spec = serving_spec
        self.track_models = track_models
        self.parameters = parameters
        self._graph = None
        self.graph = graph
        # Termination grace period is internal for runtimes that have a pod termination hook hence it is not in the
        # _dict_fields and doesn't have a setter.
        self._termination_grace_period_seconds = None
        self.__fields_pending_discard = {}

    @property
    def volumes(self) -> list:
        return list(self._volumes.values())

    @volumes.setter
    def volumes(self, volumes):
        self._volumes = {}
        if volumes:
            for vol in volumes:
                set_named_item(self._volumes, vol)

    @property
    def volume_mounts(self) -> list:
        return list(self._volume_mounts.values())

    @volume_mounts.setter
    def volume_mounts(self, volume_mounts):
        self._volume_mounts = {}
        if volume_mounts:
            for volume_mount in volume_mounts:
                self._set_volume_mount(volume_mount)

    @property
    def affinity(self) -> k8s_client.V1Affinity:
        return self._affinity

    @affinity.setter
    def affinity(self, affinity):
        self._affinity = transform_attribute_to_k8s_class_instance("affinity", affinity)

    @property
    def tolerations(self) -> list[k8s_client.V1Toleration]:
        return self._tolerations

    @tolerations.setter
    def tolerations(self, tolerations):
        self._tolerations = transform_attribute_to_k8s_class_instance(
            "tolerations", tolerations
        )

    @property
    def resources(self) -> dict:
        return self._resources

    @resources.setter
    def resources(self, resources):
        self._resources = self.enrich_resources_with_default_pod_resources(
            "resources", resources
        )

    @property
    def preemption_mode(self) -> str:
        return self._preemption_mode

    @preemption_mode.setter
    def preemption_mode(self, mode):
        self._preemption_mode = mode or mlconf.function_defaults.preemption_mode

    @property
    def security_context(self) -> k8s_client.V1SecurityContext:
        return self._security_context

    @security_context.setter
    def security_context(self, security_context):
        self._security_context = transform_attribute_to_k8s_class_instance(
            "security_context", security_context
        )

    @property
    def termination_grace_period_seconds(self) -> typing.Optional[int]:
        return self._termination_grace_period_seconds

    @property
    def graph(self):
        """states graph, holding the serving workflow/DAG topology"""
        return self._graph

    @graph.setter
    def graph(self, graph):
        from ..serving.states import graph_root_setter

        graph_root_setter(self, graph)

    def _serialize_field(
        self, struct: dict, field_name: typing.Optional[str] = None, strip: bool = False
    ) -> typing.Any:
        """
        Serialize a field to a dict, list, or primitive type.
        If field_name is in _k8s_fields_to_serialize, we will apply k8s serialization
        """
        k8s_api = k8s_client.ApiClient()
        if field_name in self._k8s_fields_to_serialize:
            return k8s_api.sanitize_for_serialization(getattr(self, field_name))
        return super()._serialize_field(struct, field_name, strip)

    def _enrich_field(
        self, struct: dict, field_name: typing.Optional[str] = None, strip: bool = False
    ) -> typing.Any:
        k8s_api = k8s_client.ApiClient()
        if strip:
            if field_name == "env":
                # We first try to pull from struct because the field might have been already serialized and if not,
                # we pull from self
                envs = struct.get(field_name, None) or getattr(self, field_name, None)
                if envs:
                    serialized_envs = k8s_api.sanitize_for_serialization(envs)
                    for env in serialized_envs:
                        if env["name"].startswith("V3IO_"):
                            env["value"] = ""
                    return serialized_envs
        return super()._enrich_field(struct=struct, field_name=field_name, strip=strip)

    def _apply_enrichment_before_to_dict_completion(
        self, struct: dict, strip: bool = False
    ):
        if strip:
            # Reset this, since mounts and env variables were cleared.
            struct["disable_auto_mount"] = False
        return super()._apply_enrichment_before_to_dict_completion(struct, strip)

    def update_vols_and_mounts(
        self, volumes, volume_mounts, volume_mounts_field_name="_volume_mounts"
    ):
        if volumes:
            for vol in volumes:
                set_named_item(self._volumes, vol)

        if volume_mounts:
            for volume_mount in volume_mounts:
                self._set_volume_mount(volume_mount, volume_mounts_field_name)

    def validate_service_account(self, allowed_service_accounts):
        if (
            allowed_service_accounts
            and self.service_account not in allowed_service_accounts
        ):
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Function service account {self.service_account} is not in allowed "
                + f"service accounts {allowed_service_accounts}"
            )

    def with_volumes(
        self,
        volumes: typing.Union[list[dict], dict, V1Volume],
    ) -> "KubeResourceSpec":
        """Add volumes to the volumes dictionary, only used as part of the mlrun_pipelines mount functions."""
        if isinstance(volumes, dict):
            set_named_item(self._volumes, volumes)
        elif isinstance(volumes, Iterable):
            for volume in volumes:
                set_named_item(self._volumes, volume)
        else:
            set_named_item(self._volumes, volumes)
        return self

    def with_volume_mounts(
        self,
        volume_mounts: typing.Union[list[dict], dict, V1VolumeMount],
    ) -> "KubeResourceSpec":
        """Add volume mounts to the volume mounts dictionary,
        only used as part of the mlrun_pipelines mount functions."""
        if isinstance(volume_mounts, dict):
            self._set_volume_mount(volume_mounts)
        elif isinstance(volume_mounts, Iterable):
            for volume_mount in volume_mounts:
                self._set_volume_mount(volume_mount)
        else:
            self._set_volume_mount(volume_mounts)
        return self

    def _set_volume_mount(
        self, volume_mount, volume_mounts_field_name="_volume_mounts"
    ):
        # using the mountPath as the key cause it must be unique (k8s limitation)
        # volume_mount may be an V1VolumeMount instance (object access, snake case) or sanitized dict (dict
        # access, camel case)
        getattr(self, volume_mounts_field_name)[
            get_item_name(volume_mount, "mountPath")
            or get_item_name(volume_mount, "mount_path")
        ] = volume_mount

    def _verify_and_set_limits(
        self,
        resources_field_name,
        mem: typing.Optional[str] = None,
        cpu: typing.Optional[str] = None,
        gpus: typing.Optional[int] = None,
        gpu_type: str = "nvidia.com/gpu",
        patch: bool = False,
    ):
        resources = verify_limits(
            resources_field_name, mem=mem, cpu=cpu, gpus=gpus, gpu_type=gpu_type
        )
        if not patch:
            update_in(
                getattr(self, resources_field_name),
                "limits",
                resources,
            )
        else:
            for resource, resource_value in resources.items():
                # gpu_type can contain "." (e.g nvidia.com/gpu) in the name which will result the `update_in` to split
                # the resource name
                if resource == gpu_type:
                    limits: dict = getattr(self, resources_field_name).setdefault(
                        "limits", {}
                    )
                    limits.update({resource: resource_value})
                else:
                    update_in(
                        getattr(self, resources_field_name),
                        f"limits.{resource}",
                        resource_value,
                    )

    def discard_changes(self):
        """
        Certain pipeline engines might make temporary changes to a function spec to ensure expected behavior.
        Kubeflow, for instance, can use pipeline parameters to change resource requests/limits on a function.
        Affected fields will be scheduled for cleanup automatically. Direct user intervention is not required.
        """
        for k, v in self.__fields_pending_discard.items():
            setattr(self, k, v)

    def _add_field_to_pending_discard(self, field_name, field_value):
        self.__fields_pending_discard.update(
            {
                field_name: copy.deepcopy(field_value),
            }
        )

    def _verify_and_set_requests(
        self,
        resources_field_name,
        mem: typing.Optional[str] = None,
        cpu: typing.Optional[str] = None,
        patch: bool = False,
    ):
        resources = verify_requests(resources_field_name, mem=mem, cpu=cpu)
        for pattern in mlrun.utils.regex.pipeline_param:
            if re.match(pattern, str(cpu)) or re.match(pattern, str(mem)):
                self._add_field_to_pending_discard(
                    resources_field_name, getattr(self, resources_field_name)
                )
        if not patch:
            update_in(
                getattr(self, resources_field_name),
                "requests",
                resources,
            )
        else:
            for resource, resource_value in resources.items():
                update_in(
                    getattr(self, resources_field_name),
                    f"requests.{resource}",
                    resource_value,
                )

    def with_limits(
        self,
        mem: typing.Optional[str] = None,
        cpu: typing.Optional[str] = None,
        gpus: typing.Optional[int] = None,
        gpu_type: str = "nvidia.com/gpu",
        patch: bool = False,
    ):
        """
        Set pod cpu/memory/gpu limits (max values)

        :param mem:     set limit for memory e.g. '500M', '2G', etc.
        :param cpu:     set limit for cpu e.g. '0.5', '2', etc.
        :param gpus:    set limit for gpu
        :param gpu_type:    set gpu type e.g. "nvidia.com/gpu"
        :param patch:    by default it overrides the whole limits section,
                        if you wish to patch specific resources use `patch=True`
        """
        self._verify_and_set_limits("resources", mem, cpu, gpus, gpu_type, patch=patch)

    def with_requests(
        self,
        mem: typing.Optional[str] = None,
        cpu: typing.Optional[str] = None,
        patch: bool = False,
    ):
        """
        Set requested (desired) pod cpu/memory resources

        :param mem:     set request for memory e.g. '200M', '1G', etc.
        :param cpu:     set request for cpu e.g. '0.1', '1', etc.
        :param patch:   by default it overrides the whole requests section,
                        if you wish to patch specific resources use `patch=True`
        """
        self._verify_and_set_requests("resources", mem, cpu, patch)

    def enrich_resources_with_default_pod_resources(
        self, resources_field_name: str, resources: dict
    ):
        resources_types = ["cpu", "memory"]
        resource_requirements = ["requests", "limits"]
        default_resources = mlconf.get_default_function_pod_resources()

        if resources:
            for resource_requirement in resource_requirements:
                for resource_type in resources_types:
                    if (
                        resources.setdefault(resource_requirement, {}).setdefault(
                            resource_type
                        )
                        is None
                    ):
                        resources[resource_requirement][resource_type] = (
                            default_resources[resource_requirement][resource_type]
                        )
        # This enables the user to define that no defaults would be applied on the resources
        elif resources == {}:
            return resources
        else:
            resources = default_resources
        resources["requests"] = verify_requests(
            resources_field_name,
            mem=resources["requests"]["memory"],
            cpu=resources["requests"]["cpu"],
        )
        gpu_type, gpu_value = get_gpu_from_resource_requirement(resources["limits"])
        resources["limits"] = verify_limits(
            resources_field_name,
            mem=resources["limits"]["memory"],
            cpu=resources["limits"]["cpu"],
            gpus=gpu_value,
            gpu_type=gpu_type,
        )
        if not resources["requests"] and not resources["limits"]:
            return {}
        return resources

    @staticmethod
    def _prune_node_selector_requirements_from_node_selector_terms(
        node_selector_terms: list[k8s_client.V1NodeSelectorTerm],
        node_selector_requirements_to_prune: list[k8s_client.V1NodeSelectorRequirement],
    ) -> list[k8s_client.V1NodeSelectorTerm]:
        """
        Goes over each expression in all the terms provided and removes the expressions if it matches
        one of the requirements provided to remove

        :return: New list of terms without the provided node selector requirements
        """
        new_node_selector_terms: list[k8s_client.V1NodeSelectorTerm] = []
        for term in node_selector_terms:
            new_node_selector_requirements: list[
                k8s_client.V1NodeSelectorRequirement
            ] = []
            for node_selector_requirement in term.match_expressions:
                to_prune = False
                # go over each requirement and check if matches the current expression
                for (
                    node_selector_requirement_to_prune
                ) in node_selector_requirements_to_prune:
                    if node_selector_requirement == node_selector_requirement_to_prune:
                        to_prune = True
                        # no need to keep going over the list provided for the current expression
                        break
                if not to_prune:
                    new_node_selector_requirements.append(node_selector_requirement)

            # check if there is something to add
            if len(new_node_selector_requirements) > 0 or term.match_fields:
                # Add new node selector terms without the matching expressions to prune
                new_node_selector_terms.append(
                    k8s_client.V1NodeSelectorTerm(
                        match_expressions=new_node_selector_requirements,
                        match_fields=term.match_fields,
                    )
                )
        return new_node_selector_terms


class AutoMountType(str, Enum):
    none = "none"
    auto = "auto"
    v3io_credentials = "v3io_credentials"
    v3io_fuse = "v3io_fuse"
    pvc = "pvc"
    s3 = "s3"
    env = "env"

    @classmethod
    def _missing_(cls, value):
        if value:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Invalid value for auto_mount_type - '{value}'"
            )
        return AutoMountType.default()

    @staticmethod
    def default():
        return AutoMountType.auto

    # Any modifier that configures a mount on a runtime should be included here. These modifiers, if applied to the
    # runtime, will suppress the auto-mount functionality.
    @classmethod
    def all_mount_modifiers(cls):
        return [
            mlrun.runtimes.mounts.v3io_cred.__name__,
            mlrun.runtimes.mounts.mount_v3io.__name__,
            mlrun.runtimes.mounts.mount_pvc.__name__,
            mlrun.runtimes.mounts.auto_mount.__name__,
            mlrun.runtimes.mounts.mount_s3.__name__,
            mlrun.runtimes.mounts.set_env_variables.__name__,
        ]

    @classmethod
    def is_auto_modifier(cls, modifier):
        # Check if modifier is one of the known mount modifiers. We need to use startswith since the modifier itself is
        # a nested function returned from the modifier function (such as 'v3io_cred.<locals>._use_v3io_cred')
        modifier_name = modifier.__qualname__
        return any(
            modifier_name.startswith(mount_modifier)
            for mount_modifier in AutoMountType.all_mount_modifiers()
        )

    @staticmethod
    def _get_auto_modifier():
        # If we're running on Iguazio - use v3io_cred
        if mlconf.igz_version != "":
            return mlrun.runtimes.mounts.v3io_cred
        # Else, either pvc mount if it's configured or do nothing otherwise
        pvc_configured = (
            "MLRUN_PVC_MOUNT" in os.environ
            or "pvc_name" in mlconf.get_storage_auto_mount_params()
        )
        return mlrun.runtimes.mounts.mount_pvc if pvc_configured else None

    def get_modifier(self):
        return {
            AutoMountType.none: None,
            AutoMountType.v3io_credentials: mlrun.runtimes.mounts.v3io_cred,
            AutoMountType.v3io_fuse: mlrun.runtimes.mounts.mount_v3io,
            AutoMountType.pvc: mlrun.runtimes.mounts.mount_pvc,
            AutoMountType.auto: self._get_auto_modifier(),
            AutoMountType.s3: mlrun.runtimes.mounts.mount_s3,
            AutoMountType.env: mlrun.runtimes.mounts.set_env_variables,
        }[self]


class KubeResource(BaseRuntime):
    """
    A parent class for runtimes that generate k8s resources when executing.
    """

    kind = "job"
    _is_nested = True

    def __init__(self, spec=None, metadata=None):
        super().__init__(metadata=metadata, spec=spec)
        self.verbose = False

    @property
    def spec(self) -> KubeResourceSpec:
        return self._spec

    @spec.setter
    def spec(self, spec):
        self._spec = self._verify_dict(spec, "spec", KubeResourceSpec)

    def set_env_from_secret(
        self,
        name: str,
        secret: Optional[str] = None,
        secret_key: Optional[str] = None,
    ):
        """
        Set an environment variable from a Kubernetes Secret.
        Client-side guard forbids MLRun internal auth/project secrets; no-op on API.
        """
        mlrun.common.secrets.validate_not_forbidden_secret(secret)
        key = secret_key or name
        value_from = k8s_client.V1EnvVarSource(
            secret_key_ref=k8s_client.V1SecretKeySelector(name=secret, key=key)
        )
        return self._set_env(name=name, value_from=value_from)

    def set_env(
        self,
        name: str,
        value: Optional[str] = None,
        value_from: Optional[typing.Any] = None,
    ):
        """
        Set an environment variable.
        If value comes from a Secret, validate on client-side only.
        """
        if value_from is not None:
            secret_name = self._extract_secret_name_from_value_from(
                value_from=value_from
            )
            if secret_name:
                mlrun.common.secrets.validate_not_forbidden_secret(secret_name)
            return self._set_env(name=name, value_from=value_from)

        # Plain literal value path
        return self._set_env(
            name=name, value=(str(value) if value is not None else None)
        )

    def with_annotations(self, annotations: dict):
        """set a key/value annotations in the metadata of the pod"""
        for key, value in annotations.items():
            self.metadata.annotations[key] = str(value)
        return self

    def get_env(self, name, default=None):
        """Get the pod environment variable for the given name, if not found return the default
        If it's a scalar value, will return it, if the value is from source, return the k8s struct (V1EnvVarSource)
        """
        for env_var in self.spec.env:
            if get_item_name(env_var) == name:
                # valueFrom is a workaround for now, for some reason the envs aren't getting sanitized
                # TODO: add env to sanitized attributes and then remove the valueFrom as the sanitized env will have
                #   value_from key and not valueFrom
                for value_key in ["value", "value_from", "valueFrom"]:
                    value = get_item_name(env_var, value_key)
                    if value is not None:
                        return value
        return default

    def is_env_exists(self, name):
        """Check whether there is an environment variable define for the given key"""
        for env_var in self.spec.env:
            if get_item_name(env_var) == name:
                return True
        return False

    def _set_env(self, name, value=None, value_from=None):
        new_var = k8s_client.V1EnvVar(name=name, value=value, value_from=value_from)

        # ensure we don't have duplicate env vars with the same name
        for env_index, value_item in enumerate(self.spec.env):
            if get_item_name(value_item) == name:
                self.spec.env[env_index] = new_var
                return self
        self.spec.env.append(new_var)
        return self

    def set_envs(
        self,
        env_vars: typing.Optional[dict] = None,
        file_path: typing.Optional[str] = None,
    ):
        """set pod environment var from key/value dict or .env file

        :param env_vars:  dict with env key/values
        :param file_path: .env file with key=value lines
        """
        if (not env_vars and not file_path) or (env_vars and file_path):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "must specify env_vars OR file_path"
            )
        if file_path:
            if os.path.isfile(file_path):
                env_vars = dotenv.dotenv_values(file_path)
                if None in env_vars.values():
                    raise mlrun.errors.MLRunInvalidArgumentError(
                        "env file lines must be in the form key=value"
                    )
            else:
                raise mlrun.errors.MLRunNotFoundError(f"{file_path} does not exist")
        for name, value in env_vars.items():
            if isinstance(value, dict) and "valueFrom" in value:
                self.set_env(name, value_from=value["valueFrom"])
            else:
                self.set_env(name, value)
        return self

    def set_image_pull_configuration(
        self,
        image_pull_policy: typing.Optional[str] = None,
        image_pull_secret_name: typing.Optional[str] = None,
    ):
        """
        Configure the image pull parameters for the runtime.

        :param image_pull_policy: The policy to use when pulling. One of `IfNotPresent`, `Always` or `Never`
        :param image_pull_secret_name: Name of a k8s secret containing image repository's authentication credentials
        """
        if image_pull_policy is not None:
            allowed_policies = ["Always", "IfNotPresent", "Never"]
            if image_pull_policy not in allowed_policies:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"Image pull policy must be one of {allowed_policies}, got {image_pull_policy}"
                )
            self.spec.image_pull_policy = image_pull_policy
        if image_pull_secret_name is not None:
            self.spec.image_pull_secret = image_pull_secret_name

    def set_state_thresholds(
        self,
        state_thresholds: dict[str, str],
        patch: bool = True,
    ):
        """
        Set the threshold for a specific state of the runtime.
        The threshold is the amount of time that the runtime will wait before aborting the run if the job is in the
        matching state.
        The threshold time string must conform to timelength python package standards and be at least 1 minute
        (e.g. 1000s, 1 hour 30m, 1h etc. or -1 for infinite).
        If the threshold is not set for a state, the default threshold will be used.

        :param state_thresholds: A dictionary of state to threshold. The supported states are:

                                 * pending_scheduled - The pod/crd is scheduled on a node but not yet running
                                 * pending_not_scheduled - The pod/crd is not yet scheduled on a node
                                 * executing - The pod/crd started and is running
                                 * image_pull_backoff - The pod/crd is in image pull backoff

                                See :code:`mlrun.mlconf.function.spec.state_thresholds` for the default thresholds.
        :param patch: Whether to merge the given thresholds with the existing thresholds (True, default)
                      or override them (False)
        """
        if patch:
            self.spec.state_thresholds.update(state_thresholds)
        else:
            self.spec.state_thresholds = state_thresholds

    def with_limits(
        self,
        mem: typing.Optional[str] = None,
        cpu: typing.Optional[str] = None,
        gpus: typing.Optional[int] = None,
        gpu_type: str = "nvidia.com/gpu",
        patch: bool = False,
    ):
        """
        Set pod cpu/memory/gpu limits (max values)

        :param mem:     set limit for memory e.g. '500M', '2G', etc.
        :param cpu:     set limit for cpu e.g. '0.5', '2', etc.
        :param gpus:    set limit for gpu
        :param gpu_type:    set gpu type e.g. "nvidia.com/gpu"
        :param patch:    by default it overrides the whole limits section,
                        if you wish to patch specific resources use `patch=True`
        """
        self.spec.with_limits(mem, cpu, gpus, gpu_type, patch=patch)

    def with_requests(
        self,
        mem: typing.Optional[str] = None,
        cpu: typing.Optional[str] = None,
        patch: bool = False,
    ):
        """
        Set requested (desired) pod cpu/memory resources

        :param mem:     set request for memory e.g. '200M', '1G', etc.
        :param cpu:     set request for cpu e.g. '0.1', '1', etc.
        :param patch:   by default it overrides the whole requests section,
                        if you wish to patch specific resources use `patch=True`
        """
        self.spec.with_requests(mem, cpu, patch=patch)

    @staticmethod
    def detect_preemptible_node_selector(node_selector: dict[str, str]) -> list[str]:
        """
        Check whether any provided node selector matches preemptible selectors.

        :param node_selector: User-provided node selector mapping.
        :return: List of `"key='value'"` strings that match a preemptible selector.
        """
        preemptible_node_selector = mlconf.get_preemptible_node_selector()

        return [
            f"'{key}': '{val}'"
            for key, val in node_selector.items()
            if preemptible_node_selector.get(key) == val
        ]

    def detect_preemptible_tolerations(
        self, tolerations: list[k8s_client.V1Toleration]
    ) -> list[str]:
        """
        Check whether any provided toleration matches preemptible tolerations.

        :param tolerations: User-provided tolerations.
        :return: List of formatted toleration strings that are considered preemptible.
        """
        preemptible_tolerations = [
            k8s_client.V1Toleration(
                key=toleration.get("key"),
                value=toleration.get("value"),
                effect=toleration.get("effect"),
            )
            for toleration in mlconf.get_preemptible_tolerations()
        ]

        def _format_toleration(toleration):
            return f"'{toleration.key}'='{toleration.value}' (effect: '{toleration.effect}')"

        return [
            _format_toleration(toleration)
            for toleration in tolerations
            if toleration in preemptible_tolerations
        ]

    def detect_preemptible_affinity(self, affinity: k8s_client.V1Affinity) -> list[str]:
        """
        Check whether any provided affinity rules match preemptible affinity configs.

        :param affinity: User-provided affinity object.
        :return: List of formatted expressions that overlap with preemptible terms.
        """
        preemptible_affinity_terms = generate_preemptible_nodes_affinity_terms()
        conflicting_affinities = []

        if (
            affinity
            and affinity.node_affinity
            and affinity.node_affinity.required_during_scheduling_ignored_during_execution
        ):
            user_terms = affinity.node_affinity.required_during_scheduling_ignored_during_execution.node_selector_terms
            for user_term in user_terms:
                user_expressions = {
                    (expr.key, expr.operator, tuple(expr.values or []))
                    for expr in user_term.match_expressions or []
                }

                for preemptible_term in preemptible_affinity_terms:
                    preemptible_expressions = {
                        (expr.key, expr.operator, tuple(expr.values or []))
                        for expr in preemptible_term.match_expressions or []
                    }

                    # Ensure operators match and preemptible expressions are present
                    common_exprs = user_expressions & preemptible_expressions
                    if common_exprs:
                        formatted = ", ".join(
                            f"'{key}  {operator}  {list(values)}'"
                            for key, operator, values in common_exprs
                        )
                        conflicting_affinities.append(formatted)
        return conflicting_affinities

    def raise_preemptible_warning(
        self,
        node_selector: typing.Optional[dict[str, str]],
        tolerations: typing.Optional[list[k8s_client.V1Toleration]],
        affinity: typing.Optional[k8s_client.V1Affinity],
    ) -> None:
        """
        Detect conflicts and emit a single consolidated warning if needed.

        :param node_selector: User-provided node selector.
        :param tolerations: User-provided tolerations.
        :param affinity: User-provided affinity.
        :warns: PreemptionWarning - Emitted when any of the provided selectors,
                tolerations, or affinity terms match the configured preemptible
                settings. The message lists the conflicting items.
        """
        conflict_messages = []

        if node_selector:
            ns_conflicts = ", ".join(
                self.detect_preemptible_node_selector(node_selector)
            )
            if ns_conflicts:
                conflict_messages.append(f"Node selectors: {ns_conflicts}")

        if tolerations:
            tol_conflicts = ", ".join(self.detect_preemptible_tolerations(tolerations))
            if tol_conflicts:
                conflict_messages.append(f"Tolerations: {tol_conflicts}")

        if affinity:
            affinity_conflicts = ", ".join(self.detect_preemptible_affinity(affinity))
            if affinity_conflicts:
                conflict_messages.append(f"Affinity: {affinity_conflicts}")

        if conflict_messages:
            warning_componentes = "; \n".join(conflict_messages)
            warnings.warn(
                f"Warning: based on MLRun's preemptible node configuration, the following components \n"
                f"may be removed or adjusted at runtime:\n"
                f"{warning_componentes}.\n"
                "This adjustment depends on the function's preemption mode. \n"
                "The list of potential adjusted preemptible selectors can be viewed here: "
                "mlrun.mlconf.get_preemptible_node_selector() and mlrun.mlconf.get_preemptible_tolerations()."
            )

    def with_node_selection(
        self,
        node_name: typing.Optional[str] = None,
        node_selector: typing.Optional[dict[str, str]] = None,
        affinity: typing.Optional[k8s_client.V1Affinity] = None,
        tolerations: typing.Optional[list[k8s_client.V1Toleration]] = None,
    ):
        """
        Configure Kubernetes node scheduling for this function.

        Updates one or more scheduling hints: exact node pinning, label-based selection,
        affinity/anti-affinity rules, and taint tolerations. Passing ``None`` leaves the
        current value unchanged; pass an empty dict/list (e.g., ``{}``, ``[]``) to clear.

        :param node_name: Exact Kubernetes node name to pin the pod to.
        :param node_selector: Mapping of label selectors. Use ``{}`` to clear.
        :param affinity: :class:`kubernetes.client.V1Affinity` constraints.
        :param tolerations: List of :class:`kubernetes.client.V1Toleration`. Use ``[]`` to clear.
        :warns: PreemptionWarning - Emitted if provided selectors/tolerations/affinity
                conflict with the function's preemption mode.

        Example usage:
            Prefer a GPU pool and allow scheduling on spot nodes::

                job.with_node_selection(
                    node_selector={"nodepool": "gpu"},
                    tolerations=[k8s_client.V1Toleration(key="spot", operator="Exists")],
                )
        """
        if node_name:
            self.spec.node_name = node_name
        if node_selector is not None:
            validate_node_selectors(node_selectors=node_selector, raise_on_error=False)
            self.spec.node_selector = node_selector
        if affinity is not None:
            self.spec.affinity = affinity
        if tolerations is not None:
            self.spec.tolerations = tolerations
        self.raise_preemptible_warning(
            node_selector=self.spec.node_selector,
            tolerations=self.spec.tolerations,
            affinity=self.spec.affinity,
        )

    def with_priority_class(self, name: typing.Optional[str] = None):
        """
        Enables to control the priority of the pod
        If not passed - will default to mlrun.mlconf.default_function_priority_class_name

        :param name:       The name of the priority class
        """
        if name is None:
            name = mlconf.default_function_priority_class_name
        valid_priority_class_names = self.list_valid_priority_class_names()
        if name not in valid_priority_class_names:
            message = "Priority class name not in available priority class names"
            logger.warning(
                message,
                priority_class_name=name,
                valid_priority_class_names=valid_priority_class_names,
            )
            raise mlrun.errors.MLRunInvalidArgumentError(message)
        self.spec.priority_class_name = name

    def with_preemption_mode(self, mode: typing.Union[PreemptionModes, str]):
        """
        Preemption mode controls whether pods can be scheduled on preemptible nodes.
        Tolerations, node selector, and affinity are populated on preemptible nodes corresponding to the function spec.

        The supported modes are:

        * **allow** - The function can be scheduled on preemptible nodes
        * **constrain** - The function can only run on preemptible nodes
        * **prevent** - The function cannot be scheduled on preemptible nodes
        * **none** - No preemptible configuration will be applied on the function

        The default preemption mode is configurable in mlrun.mlconf.function_defaults.preemption_mode,
        by default it's set to **prevent**

        :param mode: allow | constrain | prevent | none defined in :py:class:`~mlrun.common.schemas.PreemptionModes`
        """
        preemptible_mode = PreemptionModes(mode)
        self.spec.preemption_mode = preemptible_mode.value

    def with_security_context(self, security_context: k8s_client.V1SecurityContext):
        """
        Set security context for the pod.
        For Iguazio we handle security context internally -
        see mlrun.common.schemas.function.SecurityContextEnrichmentModes

        Example::

            from kubernetes import client as k8s_client

            security_context = k8s_client.V1SecurityContext(
                run_as_user=1000,
                run_as_group=3000,
            )
            function.with_security_context(security_context)

        More info:
        https://kubernetes.io/docs/tasks/configure-pod-container/security-context/#set-the-security-context-for-a-pod

        :param security_context:         The security context for the pod
        """
        if (
            mlrun.mlconf.function.spec.security_context.enrichment_mode
            != SecurityContextEnrichmentModes.disabled.value
        ):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Security context is handled internally when enrichment mode is not disabled"
            )
        self.spec.security_context = security_context

    def apply(
        self,
        modifier: typing.Callable[["KubeResource"], "KubeResource"],
    ) -> "KubeResource":
        """
        Apply a modifier to the runtime which is used to change the runtimes k8s object's spec.
        All modifiers accept Kube, apply some changes on its spec and return it so modifiers can be chained
        one after the other.

        :param modifier: a modifier callable object
        :return: the runtime (self) after the modifications
        """
        modifier(self)
        if AutoMountType.is_auto_modifier(modifier):
            self.spec.disable_auto_mount = True

        api_client = k8s_client.ApiClient()
        if self.spec.env:
            for index, env in enumerate(
                api_client.sanitize_for_serialization(self.spec.env)
            ):
                self.spec.env[index] = env

        if self.spec.volumes and self.spec.volume_mounts:
            vols = api_client.sanitize_for_serialization(self.spec.volumes)
            mounts = api_client.sanitize_for_serialization(self.spec.volume_mounts)
            self.spec.update_vols_and_mounts(vols, mounts)

        return self

    def list_valid_priority_class_names(self):
        return mlconf.get_valid_function_priority_class_names()

    def get_default_priority_class_name(self):
        return mlconf.default_function_priority_class_name

    def try_auto_mount_based_on_config(self, override_params=None):
        if self.spec.disable_auto_mount:
            logger.debug(
                "Mount already applied or auto-mount manually disabled - not performing auto-mount"
            )
            return

        auto_mount_type = AutoMountType(mlconf.storage.auto_mount_type)
        modifier = auto_mount_type.get_modifier()
        if not modifier:
            logger.debug(
                "Auto mount disabled due to user selection (auto_mount_type=none)"
            )
            return

        mount_params_dict = mlconf.get_storage_auto_mount_params()
        override_params = override_params or {}
        for key, value in override_params.items():
            mount_params_dict[key] = value

        mount_params_dict = _filter_modifier_params(modifier, mount_params_dict)

        self.apply(modifier(**mount_params_dict))

    def validate_and_enrich_service_account(
        self, allowed_service_accounts, default_service_account
    ):
        if not self.spec.service_account:
            if default_service_account:
                self.spec.service_account = default_service_account
                logger.info(
                    f"Setting default service account to function: {default_service_account}"
                )

        self.spec.validate_service_account(allowed_service_accounts)

    def _configure_mlrun_build_with_source(
        self, source, workdir=None, handler=None, pull_at_runtime=True, target_dir=None
    ):
        mlrun.utils.helpers.validate_builder_source(source, pull_at_runtime, workdir)

        self.spec.build.source = source
        if handler:
            self.spec.default_handler = handler
        if workdir:
            self.spec.workdir = workdir
        if target_dir:
            self.spec.build.source_code_target_dir = target_dir

        self.spec.build.load_source_on_run = pull_at_runtime
        if (
            self.spec.build.base_image
            and not self.spec.build.commands
            and pull_at_runtime
            and not self.spec.image
        ):
            # if we load source from repo and don't need a full build use the base_image as the image
            self.spec.image = self.spec.build.base_image
        elif not pull_at_runtime:
            # clear the image so build will not be skipped
            self.spec.build.base_image = self.spec.build.base_image or self.spec.image
            self.spec.image = ""

    def _resolve_build_with_mlrun(self, with_mlrun: typing.Optional[bool] = None):
        build = self.spec.build
        if with_mlrun is None:
            if build.with_mlrun is not None:
                with_mlrun = build.with_mlrun
            else:
                with_mlrun = build.base_image and not (
                    build.base_image.startswith("mlrun/")
                    or "/mlrun/" in build.base_image
                )
        if (
            not build.source
            and not build.commands
            and not build.requirements
            and not build.extra
            and with_mlrun
        ):
            logger.info(
                "Running build to add mlrun package, set "
                "with_mlrun=False to skip if its already in the image"
            )
        return with_mlrun

    def _build_image(
        self,
        builder_env: dict,
        force_build: bool,
        mlrun_version_specifier: typing.Optional[bool],
        show_on_failure: bool,
        skip_deployed: bool,
        watch: bool,
        is_kfp: bool,
        with_mlrun: typing.Optional[bool],
    ):
        # When we're in pipelines context we must watch otherwise the pipelines pod will exit before the operation
        # is actually done. (when a pipelines pod exits, the pipeline step marked as done)
        if is_kfp:
            watch = True

        if skip_deployed and self.requires_build() and not self.is_deployed():
            logger.warning(
                f"Even though {skip_deployed=}, the build might be triggered due to the function's configuration. "
                "See requires_build() and is_deployed() for reasoning."
            )

        db = self._get_db()
        data = db.remote_builder(
            self,
            with_mlrun,
            mlrun_version_specifier,
            skip_deployed,
            builder_env=builder_env,
            force_build=force_build,
        )
        self.status = data["data"].get("status", None)
        self.spec.image = mlrun.utils.get_in(
            data, "data.spec.image"
        ) or mlrun.utils.get_in(data, "data.spec.build.image")
        self.spec.build.base_image = self.spec.build.base_image or mlrun.utils.get_in(
            data, "data.spec.build.base_image"
        )
        # Get the source target dir in case it was enriched due to loading source
        self.spec.build.source_code_target_dir = mlrun.utils.get_in(
            data, "data.spec.build.source_code_target_dir"
        )
        ready = data.get("ready", False)
        if not ready:
            logger.info(
                f"Started building image: {data.get('data', {}).get('spec', {}).get('build', {}).get('image')}"
            )
        if watch and not ready:
            self.status.state = self._build_watch(
                watch=watch,
                show_on_failure=show_on_failure,
            )
            ready = self.status.state == "ready"
            if not ready:
                raise mlrun.errors.MLRunRuntimeError("Deploy failed")
        return ready

    def _build_watch(
        self,
        watch: bool = True,
        logs: bool = True,
        show_on_failure: bool = False,
    ):
        db = self._get_db()
        offset = 0
        events_offset = 0
        try:
            text, _, deploy_status_text_kind = db.get_builder_status(
                self,
                offset=0,
                logs=logs,
                events_offset=0,
            )
        except mlrun.db.RunDBError:
            raise ValueError("Function or build process not found")

        def print_log(_text):
            if _text and (
                not show_on_failure
                or self.status.state == mlrun.common.schemas.FunctionState.error
            ):
                print(_text, end="")

        print_log(text)
        if (
            deploy_status_text_kind
            == mlrun.common.constants.DeployStatusTextKind.events
        ):
            events_offset += len(text)
        else:
            offset += len(text)
        if watch:
            while self.status.state in [
                mlrun.common.schemas.FunctionState.pending,
                mlrun.common.schemas.FunctionState.running,
            ]:
                time.sleep(2)
                if show_on_failure:
                    text = ""
                    db.get_builder_status(self, offset=0, logs=False, events_offset=0)
                    if self.status.state == mlrun.common.schemas.FunctionState.error:
                        # re-read the full log on failure
                        text, _, deploy_status_text_kind = db.get_builder_status(
                            self,
                            offset=offset,
                            logs=logs,
                            events_offset=events_offset,
                        )
                else:
                    text, _, deploy_status_text_kind = db.get_builder_status(
                        self,
                        offset=offset,
                        logs=logs,
                        events_offset=events_offset,
                    )
                print_log(text)
                if (
                    deploy_status_text_kind
                    == mlrun.common.constants.DeployStatusTextKind.events
                ):
                    events_offset += len(text)
                else:
                    offset += len(text)

        return self.status.state

    @staticmethod
    def _extract_secret_name_from_value_from(
        value_from: typing.Any,
    ) -> Optional[str]:
        """Extract secret name from a V1EnvVarSource or dict representation."""
        if isinstance(value_from, k8s_client.V1EnvVarSource):
            if value_from.secret_key_ref:
                return value_from.secret_key_ref.name
        elif isinstance(value_from, dict):
            value_from = (
                value_from.get("valueFrom")
                or value_from.get("value_from")
                or value_from
            )
            secret_key_ref = (value_from or {}).get("secretKeyRef") or (
                value_from or {}
            ).get("secret_key_ref")
            if isinstance(secret_key_ref, dict):
                return secret_key_ref.get("name")
        return None


def _resolve_if_type_sanitized(attribute_name, attribute):
    attribute_config = sanitized_attributes[attribute_name]
    # heuristic - if one of the keys contains _ as part of the dict it means to_dict on the kubernetes
    # object performed, there's nothing we can do at that point to transform it to the sanitized version
    for key in attribute.keys():
        if "_" in key:
            raise mlrun.errors.MLRunInvalidArgumentTypeError(
                f"{attribute_name} must be instance of kubernetes {attribute_config.get('attribute_type_name')} class "
                f"but contains not sanitized key: {key}"
            )

    # then it's already the sanitized version
    return attribute


def transform_attribute_to_k8s_class_instance(
    attribute_name, attribute, is_sub_attr: bool = False
):
    if attribute_name not in sanitized_attributes:
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"{attribute_name} isn't in the available sanitized attributes"
        )
    attribute_config = sanitized_attributes[attribute_name]
    # initialize empty attribute type
    if attribute is None:
        return None
    if isinstance(attribute, dict):
        if _resolve_if_type_sanitized(attribute_name, attribute):
            api = k8s_client.ApiClient()
            # not ideal to use their private method, but looks like that's the only option
            # Taken from https://github.com/kubernetes-client/python/issues/977
            attribute_type = attribute_config["attribute_type"]
            if attribute_config["contains_many"]:
                attribute_type = attribute_config["sub_attribute_type"]
            attribute = api._ApiClient__deserialize(attribute, attribute_type)

    elif isinstance(attribute, list):
        attribute_instance = []
        for sub_attr in attribute:
            if not isinstance(sub_attr, dict):
                return attribute
            attribute_instance.append(
                transform_attribute_to_k8s_class_instance(
                    attribute_name, sub_attr, is_sub_attr=True
                )
            )
        attribute = attribute_instance
    # if user have set one attribute but its part of an attribute that contains many then return inside a list
    if (
        not is_sub_attr
        and attribute_config["contains_many"]
        and isinstance(attribute, attribute_config["sub_attribute_type"])
    ):
        # initialize attribute instance and add attribute to it,
        # mainly done when attribute is a list but user defines only sets the attribute not in the list
        attribute_instance = attribute_config["attribute_type"]()
        attribute_instance.append(attribute)
        return attribute_instance
    return attribute


def get_sanitized_attribute(spec, attribute_name: str):
    """
    When using methods like to_dict() on kubernetes class instances we're getting the attributes in snake_case
    Which is ok if we're using the kubernetes python package but not if for example we're creating CRDs that we
    apply directly. For that we need the sanitized (CamelCase) version.
    """
    attribute = getattr(spec, attribute_name)
    if attribute_name not in sanitized_attributes:
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"{attribute_name} isn't in the available sanitized attributes"
        )
    attribute_config = sanitized_attributes[attribute_name]
    if not attribute:
        return attribute_config["not_sanitized_class"]()

    # check if attribute of type dict, and then check if type is sanitized
    if isinstance(attribute, dict):
        if not isinstance(attribute_config["not_sanitized_class"], dict):
            raise mlrun.errors.MLRunInvalidArgumentTypeError(
                f"expected to be of type {attribute_config.get('not_sanitized_class')} but got dict"
            )
        if _resolve_if_type_sanitized(attribute_name, attribute):
            return attribute

    elif isinstance(attribute, list) and not isinstance(
        attribute[0], attribute_config["sub_attribute_type"]
    ):
        if not isinstance(attribute_config["not_sanitized_class"], list):
            raise mlrun.errors.MLRunInvalidArgumentTypeError(
                f"expected to be of type {attribute_config.get('not_sanitized_class')} but got list"
            )
        if _resolve_if_type_sanitized(attribute_name, attribute[0]):
            return attribute

    return sanitize_attribute(attribute)


def sanitize_attribute(attribute):
    api = k8s_client.ApiClient()
    return api.sanitize_for_serialization(attribute)


def _filter_modifier_params(modifier, params):
    # Make sure we only pass parameters that are accepted by the modifier.
    modifier_params = inspect.signature(modifier).parameters

    # If kwargs are supported by the modifier, we don't filter.
    if any(param.kind == param.VAR_KEYWORD for param in modifier_params.values()):
        return params

    param_names = modifier_params.keys()
    filtered_params = {}
    for key, value in params.items():
        if key in param_names:
            filtered_params[key] = value
        else:
            logger.warning(
                "Auto mount parameter not supported by modifier, filtered out",
                modifier=modifier.__name__,
                param=key,
            )
    return filtered_params



# Copyright 2025 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
from collections.abc import Awaitable
from typing import TYPE_CHECKING, Any, Callable, Optional, Union

import mlrun
from mlrun.datastore.model_provider.model_provider import (
    InvokeResponseFormat,
    ModelProvider,
    UsageResponseKeys,
)
from mlrun.datastore.utils import accepts_param

if TYPE_CHECKING:
    from openai._models import BaseModel  # noqa
    from openai.types.chat.chat_completion import ChatCompletion


class OpenAIProvider(ModelProvider):
    """
    OpenAIProvider is a wrapper around the OpenAI SDK that provides an interface
    for interacting with OpenAI's generative AI services.

    It supports both synchronous and asynchronous operations, allowing flexible
    integration into various workflows.

    This class extends the ModelProvider base class and implements OpenAI-specific
    functionality, including client initialization, model invocation, and custom
    operations tailored to the OpenAI API.
    """

    support_async = True
    response_class = None

    def __init__(
        self,
        parent,
        schema,
        name,
        endpoint="",
        secrets: Optional[dict] = None,
        default_invoke_kwargs: Optional[dict] = None,
    ):
        endpoint = endpoint or mlrun.mlconf.model_providers.openai_default_model
        if schema != "openai":
            raise mlrun.errors.MLRunInvalidArgumentError(
                "OpenAIProvider supports only 'openai' as the provider kind."
            )
        super().__init__(
            parent=parent,
            kind=schema,
            name=name,
            endpoint=endpoint,
            secrets=secrets,
            default_invoke_kwargs=default_invoke_kwargs,
        )
        self.options = self.get_client_options()

    @classmethod
    def _import_response_class(cls) -> None:
        if not cls.response_class:
            try:
                from openai.types.chat.chat_completion import ChatCompletion
            except ImportError as exc:
                raise ImportError("openai package is not installed") from exc
            cls.response_class = ChatCompletion

    @staticmethod
    def _extract_string_output(response: "ChatCompletion") -> str:
        """
        Extracts the text content of the first choice from an OpenAI ChatCompletion response.
        Only supports responses with a single choice. Raises an error if multiple choices exist.

        :param response: The ChatCompletion response from OpenAI.
        :return: The text content of the first message in the response.
        :raises MLRunInvalidArgumentError: If the response contains more than one choice.
        """
        if len(response.choices) != 1:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "OpenAIProvider: extracting string from response is only supported for single-response outputs"
            )
        return response.choices[0].message.content

    @classmethod
    def parse_endpoint_and_path(cls, endpoint, subpath) -> (str, str):
        if endpoint and subpath:
            endpoint = endpoint + subpath
            #  in openai there is no usage of subpath variable. if the model contains "/", it is part of the model name.
            subpath = ""
        return endpoint, subpath

    @property
    def client(self) -> Any:
        """
        Lazily return the synchronous OpenAI client.

        If the client has not been initialized yet, it will be created
        by calling `load_client`.
        """
        self.load_client()
        return self._client

    def load_client(self) -> None:
        """
        Lazily initialize the synchronous OpenAI client.

        The client is created only if it does not already exist.
        Raises ImportError if the openai package is not installed.
        """
        if self._client:
            return
        try:
            from openai import OpenAI  # noqa

            self._client = OpenAI(**self.options)
        except ImportError as exc:
            raise ImportError("openai package is not installed") from exc

    def load_async_client(self) -> None:
        """
        Lazily initialize the asynchronous OpenAI client.

        The client is created only if it does not already exist.
        Raises ImportError if the openai package is not installed.
        """
        if not self._async_client:
            try:
                from openai import AsyncOpenAI  # noqa

                self._async_client = AsyncOpenAI(**self.options)
            except ImportError as exc:
                raise ImportError("openai package is not installed") from exc

    @property
    def async_client(self) -> Any:
        """
        Return the asynchronous OpenAI client, creating it on first access.

        The client is lazily initialized via `load_async_client`.
        """
        self.load_async_client()
        return self._async_client

    def get_client_options(self) -> dict:
        res = dict(
            api_key=self._get_secret_or_env("OPENAI_API_KEY"),
            organization=self._get_secret_or_env("OPENAI_ORG_ID"),
            project=self._get_secret_or_env("OPENAI_PROJECT_ID"),
            base_url=self._get_secret_or_env("OPENAI_BASE_URL"),
            timeout=self._get_secret_or_env("OPENAI_TIMEOUT"),
            max_retries=self._get_secret_or_env("OPENAI_MAX_RETRIES"),
        )
        return self._sanitize_options(res)

    def custom_invoke(
        self, operation: Optional[Callable] = None, **invoke_kwargs
    ) -> Union["ChatCompletion", "BaseModel"]:
        """
        Invokes a model operation from the OpenAI client with the given keyword arguments.

        This method provides flexibility to either:
        - Call a specific OpenAI client operation (e.g., `client.images.generate`).
        - Default to `chat.completions.create` when no operation is provided.

        The operation must be a callable that accepts keyword arguments. If the callable
        does not accept a `model` parameter, it will be omitted from the call.

        Example:
            ```python
            result = openai_model_provider.custom_invoke(
                openai_model_provider.client.images.generate,
                prompt="A futuristic cityscape at sunset",
                n=1,
                size="1024x1024",
            )
            ```

        :param operation:       A callable representing the OpenAI operation to invoke.
                                If not provided, defaults to `client.chat.completions.create`.

        :param invoke_kwargs:   Additional keyword arguments to pass to the operation.
                                These are merged with `default_invoke_kwargs` and may
                                include parameters such as `temperature`, `max_tokens`,
                                or `messages`.

        :return:                The full response returned by the operation, typically
                                an OpenAI `ChatCompletion` or other OpenAI SDK model.
        """

        invoke_kwargs = self.get_invoke_kwargs(invoke_kwargs)
        model_kwargs = {"model": invoke_kwargs.pop("model", None) or self.model}

        if operation:
            if not callable(operation):
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "OpenAI custom_invoke operation must be a callable"
                )
            if not accepts_param(operation, "model"):
                model_kwargs = {}
            return operation(**invoke_kwargs, **model_kwargs)
        else:
            return self.client.chat.completions.create(**invoke_kwargs, **model_kwargs)

    async def async_custom_invoke(
        self,
        operation: Optional[Callable[..., Awaitable[Any]]] = None,
        **invoke_kwargs,
    ) -> Union["ChatCompletion", "BaseModel"]:
        """
        Asynchronously invokes a model operation from the OpenAI client with the given keyword arguments.

        This method provides flexibility to either:
        - Call a specific async OpenAI client operation (e.g., `async_client.images.generate`).
        - Default to `chat.completions.create` when no operation is provided.

        The operation must be an async callable that accepts keyword arguments.
        If the callable does not accept a `model` parameter, it will be omitted from the call.

        Example:
            ```python
            result = await openai_model_provider.async_custom_invoke(
                openai_model_provider.async_client.images.generate,
                prompt="A futuristic cityscape at sunset",
                n=1,
                size="1024x1024",
            )
            ```

        :param operation:       An async callable representing the OpenAI operation to invoke.
                                If not provided, defaults to `async_client.chat.completions.create`.

        :param invoke_kwargs:   Additional keyword arguments to pass to the operation.
                                These are merged with `default_invoke_kwargs` and may
                                include parameters such as `temperature`, `max_tokens`,
                                or `messages`.

        :return:                The full response returned by the awaited operation,
                                typically an OpenAI `ChatCompletion` or other OpenAI SDK model.

        """
        invoke_kwargs = self.get_invoke_kwargs(invoke_kwargs)
        model_kwargs = {"model": invoke_kwargs.pop("model", None) or self.model}
        if operation:
            if not inspect.iscoroutinefunction(operation):
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "OpenAI async_custom_invoke operation must be a coroutine function"
                )
            if not accepts_param(operation, "model"):
                model_kwargs = {}
            return await operation(**invoke_kwargs, **model_kwargs)
        else:
            return await self.async_client.chat.completions.create(
                **invoke_kwargs, **model_kwargs
            )

    def _response_handler(
        self,
        response: "ChatCompletion",
        invoke_response_format: InvokeResponseFormat = InvokeResponseFormat.FULL,
        **kwargs,
    ) -> ["ChatCompletion", str, dict[str, Any]]:
        if InvokeResponseFormat.is_str_response(invoke_response_format.value):
            str_response = self._extract_string_output(response)
            if invoke_response_format == InvokeResponseFormat.STRING:
                return str_response
            if invoke_response_format == InvokeResponseFormat.USAGE:
                usage = response.to_dict()["usage"]
                response = {
                    UsageResponseKeys.ANSWER: str_response,
                    UsageResponseKeys.USAGE: usage,
                }
        return response

    def invoke(
        self,
        messages: list[dict],
        invoke_response_format: InvokeResponseFormat = InvokeResponseFormat.FULL,
        **invoke_kwargs,
    ) -> Union[dict[str, Any], str, "ChatCompletion"]:
        """
        OpenAI-specific implementation of `ModelProvider.invoke`.
        Invokes an OpenAI model operation using the synchronous client.

        :param messages:
            A list of dictionaries representing the conversation history or input messages.
            Each dictionary should follow the format::
                {
                    "role": "system" | "user" | "assistant",
                    "content": "Message content as a string",
                }

            Example:

            .. code-block:: json

                [
                    {"role": "system", "content": "You are a helpful assistant."},
                    {"role": "user", "content": "What is the capital of France?"}
                ]

            Defaults to None if no messages are provided.

        :param invoke_response_format:
            Specifies the format of the returned response. Options:

            - "string": Returns only the generated text content, taken from a single response.
            - "usage": Combines the generated text with metadata (e.g., token usage), returning a dictionary::

                .. code-block:: json
                   {
                       "answer": "<generated_text>",
                       "usage": <ChatCompletion>.to_dict()["usage"]
                   }

            - "full": Returns the full OpenAI `ChatCompletion` object.

        :param invoke_kwargs:
            Additional keyword arguments passed to the OpenAI client.

        :return:
            A string, dictionary, or `ChatCompletion` object, depending on `invoke_response_format`.
        """

        response = self.custom_invoke(messages=messages, **invoke_kwargs)
        return self._response_handler(
            messages=messages,
            invoke_response_format=invoke_response_format,
            response=response,
        )

    async def async_invoke(
        self,
        messages: list[dict],
        invoke_response_format=InvokeResponseFormat.FULL,
        **invoke_kwargs,
    ) -> Union[str, "ChatCompletion", dict]:
        """
        OpenAI-specific implementation of `ModelProvider.async_invoke`.
        Invokes an OpenAI model operation using the asynchronous client.

        :param messages:
            A list of dictionaries representing the conversation history or input messages.
            Each dictionary should follow the format::
                {
                    "role": "system" | "user" | "assistant",
                    "content": "Message content as a string",
                }

            Example:

            .. code-block:: json

                [
                    {"role": "system", "content": "You are a helpful assistant."},
                    {"role": "user", "content": "What is the capital of France?"}
                ]

            Defaults to None if no messages are provided.

        :param invoke_response_format:
            Specifies the format of the returned response. Options:

            - "string": Returns only the generated text content, taken from a single response.
            - "usage": Combines the generated text with metadata (e.g., token usage), returning a dictionary::

                .. code-block:: json
                   {
                       "answer": "<generated_text>",
                       "usage": <ChatCompletion>.to_dict()["usage"]
                   }

            - "full": Returns the full OpenAI `ChatCompletion` object.

        :param invoke_kwargs:
            Additional keyword arguments passed to the OpenAI client.

        :return:
            A string, dictionary, or `ChatCompletion` object, depending on `invoke_response_format`.
        """
        response = await self.async_custom_invoke(messages=messages, **invoke_kwargs)
        return self._response_handler(
            messages=messages,
            invoke_response_format=invoke_response_format,
            response=response,
        )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from typing import Optional, Union

import numpy as np
import plotly.graph_objects as go

import mlrun
from mlrun.artifacts import Artifact, PlotlyArtifact

from ..._common import LoggingMode
from ..model_handler import DLModelHandler
from ..utils import DLTypes
from .logger import Logger


class MLRunLogger(Logger):
    """
    MLRun logger is logging the information collected during training / evaluation of the base logger and logging it to
    MLRun using a MLRun context. The logging includes:

    * For each epoch:

      * Tracking table: epoch, static hyperparameters, dynamic hyperparameters, training metrics, validation metrics.
      * Per iteration (batch) chart artifacts for the training and validation metrics.

    * At the end of the run:

      * Per epoch chart artifacts for the validation summaries and dynamic hyperparameters.
      * Model is logged with all the files and artifacts.
    """

    class _Loops:
        """
        The types of loops performed in a training / evaluation process. Used as names and prefixes to add to the
        metrics names when logging.
        """

        TRAINING = "training"
        VALIDATION = "validation"
        EVALUATION = "evaluation"

    def __init__(
        self,
        context: mlrun.MLClientCtx,
    ):
        """
        Initialize the MLRun logging interface to work with the given context.

        :param context:              MLRun context to log to. The context parameters can be logged as static
                                     hyperparameters.
        """
        super().__init__(context=context)

        # Prepare the artifacts collection:
        self._artifacts = {}  # type: Dict[str, Artifact]

    def log_epoch_to_context(
        self,
        epoch: int,
    ):
        """
        Log the last epoch. The last epoch information recorded in the given tracking dictionaries will be logged,
        meaning the epoch index will not be taken from the given 'epoch' parameter, but the '-1' index will be used in
        all the dictionaries. Each epoch will log the following information:

        * Results table:

          * Static hyperparameters.
          * Dynamic hyperparameters.
          * Last iteration recorded training results for loss and metrics.
          * Validation results summaries for loss and metrics.

        * Plot artifacts:

          * A chart for each of the metrics iteration results in training.
          * A chart for each of the metrics iteration results in validation.

        :param epoch: The epoch number that has just ended.
        """
        # Log the collected hyperparameters and values as results (the most recent value collected (-1 index)):
        for static_parameter, value in self._static_hyperparameters.items():
            self._context.log_result(static_parameter, value)
        if self._mode == LoggingMode.TRAINING:
            for dynamic_parameter, values in self._dynamic_hyperparameters.items():
                self._context.log_result(dynamic_parameter, values[-1])
            for metric, results in self._training_summaries.items():
                self._context.log_result(
                    f"{self._Loops.TRAINING}_{metric}", results[-1]
                )
        for metric, results in self._validation_summaries.items():
            self._context.log_result(
                f"{self._Loops.EVALUATION}_{metric}"
                if self._mode == LoggingMode.EVALUATION
                else f"{self._Loops.VALIDATION}_{metric}",
                results[-1],
            )

        # Log the epochs metrics results as chart artifacts:
        loops = (
            [self._Loops.EVALUATION]
            if self._mode == LoggingMode.EVALUATION
            else [self._Loops.TRAINING, self._Loops.VALIDATION]
        )
        metrics_dictionaries = (
            [self._validation_results]
            if self._mode == LoggingMode.EVALUATION
            else [self._training_results, self._validation_results]
        )
        for loop, metrics_dictionary in zip(loops, metrics_dictionaries):
            for metric_name in metrics_dictionary:
                # Create the plotly artifact:
                artifact = self._generate_metric_results_artifact(
                    loop=loop,
                    name=metric_name,
                    epochs_results=metrics_dictionary[metric_name],
                )
                # Log the artifact:
                self._context.log_artifact(
                    artifact,
                    local_path=artifact.key,
                    artifact_path=self._context.artifact_path,
                )
                # Collect it for later adding it to the model logging as extra data:
                self._artifacts[artifact.key.split(".")[0]] = artifact

        # Commit and commit children for MLRun flag bug:
        self._context.commit(completed=False)

    def log_run(
        self,
        model_handler: DLModelHandler,
        tag: str = "",
        labels: Optional[dict[str, DLTypes.TrackableType]] = None,
        parameters: Optional[dict[str, DLTypes.TrackableType]] = None,
        extra_data: Optional[dict[str, Union[DLTypes.TrackableType, Artifact]]] = None,
    ):
        """
        Log the run, summarizing the validation metrics and dynamic hyperparameters across all epochs. If 'update' is
        False, the collected logs will be updated to the model inside the given handler, otherwise the model will be
        saved and logged as a new artifact. The run log information will be the following:

        * Plot artifacts:

          * A chart for each of the metrics epochs results summaries across all the run (training and validation).
          * A chart for each of the dynamic hyperparameters epochs values across all the run.

        * Model artifact (only in training mode): The model will be saved and logged with all the collected artifacts
                                                  of this logger.

        :param model_handler: The model handler object holding the model to save and log.
        :param tag:           Version tag to give the logged model.
        :param labels:        Labels to log with the model.
        :param parameters:    Parameters to log with the model.
        :param extra_data:    Extra data to log with the model.
        """
        # If in training mode, log the summaries and hyperparameters artifacts:
        if self._mode == LoggingMode.TRAINING:
            # Create chart artifacts for summaries:
            for metric_name in self._training_summaries:
                # Create the plotly artifact:
                artifact = self._generate_summary_results_artifact(
                    name=metric_name,
                    training_results=self._training_summaries[metric_name],
                    validation_results=self._validation_summaries.get(
                        metric_name, None
                    ),
                )
                # Log the artifact:
                self._context.log_artifact(
                    artifact,
                    local_path=artifact.key,
                )
                # Collect it for later adding it to the model logging as extra data:
                self._artifacts[artifact.key.split(".")[0]] = artifact
            # Create chart artifacts for dynamic hyperparameters:
            for parameter_name in self._dynamic_hyperparameters:
                # Create the chart artifact:
                artifact = self._generate_dynamic_hyperparameter_values_artifact(
                    name=parameter_name,
                    values=self._dynamic_hyperparameters[parameter_name],
                )
                # Log the artifact:
                self._context.log_artifact(
                    artifact,
                    local_path=artifact.key,
                )
                # Collect it for later adding it to the model logging as extra data:
                self._artifacts[artifact.key.split(".")[0]] = artifact

        # Get the final metrics summary:
        metrics = self._generate_metrics_summary()

        # Log or update:
        model_handler.set_context(context=self._context)
        if self._mode == LoggingMode.EVALUATION:
            model_handler.update(
                labels=labels,
                parameters=parameters,
                metrics=metrics,
                extra_data=extra_data,
                artifacts=self._artifacts,
            )
        else:
            model_handler.log(
                tag=tag,
                labels=labels,
                parameters=parameters,
                metrics=metrics,
                extra_data=extra_data,
                artifacts=self._artifacts,
            )

        # Commit to update the changes, so they will be available in the MLRun UI:
        self._context.commit(completed=False)

    def _generate_metrics_summary(self) -> dict[str, float]:
        """
        Generate a metrics summary to log along the model.

        :return: The metrics summary.
        """
        # If in training mode, return both training and validation metrics:
        if self._mode == LoggingMode.TRAINING:
            return {
                **{
                    f"{self._Loops.TRAINING}_{name}": values[-1]
                    for name, values in self._training_summaries.items()
                },
                **{
                    f"{self._Loops.VALIDATION}_{name}": values[-1]
                    for name, values in self._validation_summaries.items()
                },
            }

        # Return the evaluation metrics:
        return {
            f"{self._Loops.EVALUATION}_{name}": values[-1]
            for name, values in self._validation_summaries.items()
        }

    @staticmethod
    def _generate_metric_results_artifact(
        loop: str, name: str, epochs_results: list[list[float]]
    ) -> PlotlyArtifact:
        """
        Generate a plotly artifact for the results of the metric provided.

        :param loop:           The results loop, training or validation.
        :param name:           The metric name.
        :param epochs_results: The entire metric results across all logged epochs.

        :return: The generated plotly figure wrapped in MLRun artifact.
        """
        # Parse the artifact's name:
        artifact_name = f"{loop}_{name}"

        # Initialize a plotly figure:
        metric_figure = go.Figure()

        # Add titles:
        metric_figure.update_layout(
            title=f"{loop} {name} Results",
            xaxis_title="Batches",
            yaxis_title="Results",
        )

        # Prepare the results list:
        results = list(itertools.chain(*epochs_results))

        # Prepare the epochs list:
        epochs_indices = [
            i * len(epoch_results) for i, epoch_results in enumerate(epochs_results)
        ][1:]

        # Draw:
        metric_figure.add_trace(
            go.Scatter(x=list(np.arange(len(results))), y=results, mode="lines")
        )
        for epoch_index in epochs_indices:
            metric_figure.add_vline(x=epoch_index, line_dash="dash", line_color="grey")

        # Create the plotly artifact:
        artifact = PlotlyArtifact(key=f"{artifact_name}.html", figure=metric_figure)

        return artifact

    @staticmethod
    def _generate_summary_results_artifact(
        name: str, training_results: list[float], validation_results: list[float]
    ) -> PlotlyArtifact:
        """
        Generate a plotly artifact for the results summary across all the epochs of training.

        :param name:               The metric name.
        :param training_results:   The metric training results summaries across the epochs.
        :param validation_results: The metric validation results summaries across the epochs. If validation was not
                                   performed, None should be passed.

        :return: The generated plotly figure wrapped in MLRun artifact.
        """
        # Parse the artifact's name:
        artifact_name = f"{name}_summary"

        # Initialize a plotly figure:
        summary_figure = go.Figure()

        # Add titles:
        summary_figure.update_layout(
            title=f"{name} Summary",
            xaxis_title="Epochs",
            yaxis_title="Results",
        )

        # Draw the results:
        summary_figure.add_trace(
            go.Scatter(
                x=list(np.arange(1, len(training_results) + 1)),
                y=training_results,
                mode="lines+markers",
                name="Training",
            )
        )
        if validation_results is not None:
            summary_figure.add_trace(
                go.Scatter(
                    x=list(np.arange(1, len(validation_results) + 1)),
                    y=validation_results,
                    mode="lines+markers",
                    name="Validation",
                )
            )

        # Create the plotly artifact:
        artifact = PlotlyArtifact(key=f"{artifact_name}.html", figure=summary_figure)

        return artifact

    @staticmethod
    def _generate_dynamic_hyperparameter_values_artifact(
        name: str, values: list[float]
    ) -> PlotlyArtifact:
        """
        Generate a plotly artifact for the values of the hyperparameter provided.

        :param name:   The hyperparameter name.
        :param values: The hyperparameter values across the training.

        :return: The generated plotly figure wrapped in MLRun artifact.
        """
        # Parse the artifact's name:
        artifact_name = f"{name}_values"

        # Initialize a plotly figure:
        hyperparameter_figure = go.Figure()

        # Add titles:
        hyperparameter_figure.update_layout(
            title=name,
            xaxis_title="Epochs",
            yaxis_title="Values",
        )

        # Draw the values:
        hyperparameter_figure.add_trace(
            go.Scatter(x=list(np.arange(len(values))), y=values, mode="lines+markers")
        )

        # Create the plotly artifact:
        artifact = PlotlyArtifact(
            key=f"{artifact_name}.html", figure=hyperparameter_figure
        )

        return artifact



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# import json
# import os
# from os.path import expanduser
#
# import requests
#
# from mlrun.errors import MLRunInvalidArgumentError
#
# from ..config import config as mlconf
# from ..k8s_utils import get_k8s_helper
# from .helpers import logger
#
# vault_default_prefix = "v1/secret/data"
#
#
# class VaultStore:
#     def __init__(self, token=None):
#         self._token = token
#         self.url = mlconf.secret_stores.vault.url
#
#     @property
#     def token(self):
#         if not self._token:
#             self._login()
#
#         return self._token
#
#     def _login(self):
#         if self._token:
#             return
#
#         if mlconf.secret_stores.vault.user_token != "":
#             logger.warning(
#                 "Using a user-token from configuration. This should only be done in test/debug!"
#             )
#             self._token = mlconf.secret_stores.vault.user_token
#             return
#
#         config_role = mlconf.secret_stores.vault.role
#         if config_role != "":
#             role_type, role_val = config_role.split(":", 1)
#             vault_role = f"mlrun-role-{role_type}-{role_val}"
#             self._safe_login_with_jwt_token(vault_role)
#
#         if self._token is None:
#             logger.warning(
#                 "Vault login: no vault token is available. No secrets will be accessible"
#             )
#
#     @staticmethod
#     def _generate_path(
#         prefix=vault_default_prefix,
#         user=None,
#         project=None,
#         user_prefix="users",
#         project_prefix="projects",
#     ):
#         if user and project:
#             raise MLRunInvalidArgumentError(
#                 "Both user and project were provided for Vault operations"
#             )
#
#         if user:
#             return prefix + f"/mlrun/{user_prefix}/{user}"
#         elif project:
#             return prefix + f"/mlrun/{project_prefix}/{project}"
#         else:
#             raise MLRunInvalidArgumentError(
#                 "To generate a vault secret path, either user or project must be specified"
#             )
#
#     @staticmethod
#     def _read_jwt_token():
#         # if for some reason the path to the token is not in conf, then attempt to get the SA token
#         # (works on k8s pods)
#         token_path = "/var/run/secrets/kubernetes.io/serviceaccount/token"
#         if mlconf.secret_stores.vault.token_path:
#             # Override the default SA token in case a specific token is installed in the mlconf-specified path
#             secret_token_path = expanduser(
#                 mlconf.secret_stores.vault.token_path + "/token"
#             )
#             if os.path.isfile(secret_token_path):
#                 token_path = secret_token_path
#
#         with open(token_path, "r") as token_file:
#             jwt_token = token_file.read()
#
#         return jwt_token
#
#     def _api_call(self, method, url, data=None):
#         self._login()
#
#         headers = {"X-Vault-Token": self._token}
#         full_url = self.url + "/" + url
#
#         response = requests.request(method, full_url, headers=headers, json=data)
#
#         if not response:
#             logger.error(
#                 "Vault failed the API call",
#                 status_code=response.status_code,
#                 reason=response.reason,
#                 url=url,
#             )
#         return response
#
#     # This method logins to the vault, assuming the container has a JWT token mounted as part of its assigned service
#     # account.
#     def _safe_login_with_jwt_token(self, role):
#
#         if role is None:
#             logger.warning(
#                 "login_with_token: Role passed is None. Will not attempt login"
#             )
#             return
#
#         jwt_token = self._read_jwt_token()
#
#         login_url = f"{self.url}/v1/auth/kubernetes/login"
#         data = {"jwt": jwt_token, "role": role}
#
#         response = requests.post(login_url, data=json.dumps(data))
#         if not response:
#             logger.error(
#                 "login_with_token: Vault failed the login request",
#                 role=role,
#                 status_code=response.status_code,
#                 reason=response.reason,
#             )
#             return
#         self._token = response.json()["auth"]["client_token"]
#
#     def get_secrets(self, keys, user=None, project=None):
#         secret_path = VaultStore._generate_path(user=user, project=project)
#         secrets = {}
#
#         # Since this method is called both on the client side (when constructing VaultStore before persisting to
#         # pod configuration) and on server side and in execution pods, we let this method fail gracefully in this case
#         # Should replace with something that will explode on server-side, once we have a way to do that.
#         if not self.url:
#             return secrets
#
#         response = self._api_call("GET", secret_path)
#
#         if not response:
#             return secrets
#
#         values = response.json()["data"]["data"]
#
#         # if no specific keys were asked for, return all the values available
#         if not keys:
#             return values
#
#         for key in keys:
#             if key in values:
#                 secrets[key] = values[key]
#         return secrets
#
#     def add_vault_secrets(self, items, project=None, user=None):
#         data_object = {"data": items}
#         url = VaultStore._generate_path(project=project, user=user)
#
#         response = self._api_call("POST", url, data_object)
#         if not response:
#             raise MLRunInvalidArgumentError(
#                 f"Vault failed the API call to create secrets. project={project}/user={user}"
#             )
#
#     def delete_vault_secrets(self, project=None, user=None):
#         self._login()
#         # Using the API to delete all versions + metadata of the given secret.
#         url = "v1/secret/metadata" + VaultStore._generate_path(
#             prefix="", project=project, user=user
#         )
#
#         response = self._api_call("DELETE", url)
#         if not response:
#             raise MLRunInvalidArgumentError(
#                 f"Vault failed the API call to delete secrets. project={project}/user={user}"
#             )
#
#     def create_project_policy(self, project):
#         policy_name = f"mlrun-project-{project}"
#         # TODO - need to make sure name is escaped properly and invalid chars are stripped
#         url = "v1/sys/policies/acl/" + policy_name
#
#         policy_str = (
#             f'path "secret/data/mlrun/projects/{project}" {{\n'
#             + '  capabilities = ["read", "list", "create", "delete", "update"]\n'
#             + "}\n"
#             + f'path "secret/data/mlrun/projects/{project}/*" {{\n'
#             + '  capabilities = ["read", "list", "create", "delete", "update"]\n'
#             + "}"
#         )
#
#         data_object = {"policy": policy_str}
#
#         response = self._api_call("PUT", url, data_object)
#         if not response:
#             raise MLRunInvalidArgumentError(
#                 f"Vault failed the API call to create a policy. "
#                 f"Response code: ({response.status_code}) - {response.reason}"
#             )
#         return policy_name
#
#     def create_project_role(self, project, sa, policy, namespace="default-tenant"):
#         role_name = f"mlrun-role-project-{project}"
#         # TODO - need to make sure name is escaped properly and invalid chars are stripped
#         url = "v1/auth/kubernetes/role/" + role_name
#
#         role_object = {
#             "bound_service_account_names": sa,
#             "bound_service_account_namespaces": namespace,
#             "policies": [policy],
#             "token_ttl": mlconf.secret_stores.vault.token_ttl,
#         }
#
#         response = self._api_call("POST", url, role_object)
#         if not response:
#             raise MLRunInvalidArgumentError(
#                 f"Vault failed the API call to create a secret. "
#                 f"Response code: ({response.status_code}) - {response.reason}"
#             )
#         return role_name
#
#
# def store_vault_project_secrets(project, items):
#     return VaultStore().add_vault_secrets(items, project=project)
#
#
# def add_vault_user_secrets(user, items):
#     return VaultStore().add_vault_secrets(items, user=user)
#
#
# def init_project_vault_configuration(project):
#     """Create needed configurations for this new project:
#     - Create a k8s service account with the name sa_vault_{proj name}
#     - Create a Vault policy with the name proj_{proj name}
#     - Create a Vault k8s auth role with the name role_proj_{proj name}
#     These constructs will enable any pod created as part of this project to access the project's secrets
#     in Vault, assuming that the secret which is part of the SA created is mounted to the pod.
#
#     :param project: Project name
#     """
#     logger.info("Initializing project vault configuration", project=project)
#
#     namespace = mlconf.namespace
#     k8s = get_k8s_helper(silent=True)
#     service_account_name = (
#         mlconf.secret_stores.vault.project_service_account_name.format(project=project)
#     )
#
#     secret_name = k8s.get_project_vault_secret_name(
#         project, service_account_name, namespace=namespace
#     )
#
#     if not secret_name:
#         k8s.create_project_service_account(
#             project, service_account_name, namespace=namespace
#         )
#
#     vault = VaultStore()
#     policy_name = vault.create_project_policy(project)
#     role_name = vault.create_project_role(
#         project, namespace=namespace, sa=service_account_name, policy=policy_name
#     )
#
#     logger.info("Created Vault policy. ", policy=policy_name, role=role_name)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import pathlib
from os import environ
from typing import Callable, Optional, Union

import mlrun.common.constants as mlrun_constants
import mlrun.common.schemas.schedule
import mlrun.errors
import mlrun.launcher.client as launcher
import mlrun.run
import mlrun.runtimes.generators
import mlrun.utils.clones
import mlrun.utils.notifications
from mlrun.utils import logger


class ClientLocalLauncher(launcher.ClientBaseLauncher):
    """
    ClientLocalLauncher is a launcher that runs the job locally.
    Either on the user's machine (_is_run_local is True) or on a remote machine (_is_run_local is False).
    """

    def __init__(self, local: bool = False, **kwargs):
        """
        Initialize a ClientLocalLauncher.
        :param local:   True if the job runs on the user's local machine,
                        False if it runs on a remote machine (e.g. a dedicated k8s pod).
        """
        super().__init__(**kwargs)
        self._is_run_local = local

    def launch(
        self,
        runtime: "mlrun.runtimes.BaseRuntime",
        task: Optional[
            Union["mlrun.run.RunTemplate", "mlrun.run.RunObject", dict]
        ] = None,
        handler: Optional[Union[str, Callable]] = None,
        name: Optional[str] = "",
        project: Optional[str] = "",
        params: Optional[dict] = None,
        inputs: Optional[dict[str, str]] = None,
        out_path: Optional[str] = "",
        workdir: Optional[str] = "",
        artifact_path: Optional[str] = "",
        output_path: Optional[str] = "",
        watch: Optional[bool] = True,
        schedule: Optional[
            Union[str, mlrun.common.schemas.schedule.ScheduleCronTrigger]
        ] = None,
        hyperparams: Optional[dict[str, list]] = None,
        hyper_param_options: Optional[mlrun.model.HyperParamOptions] = None,
        verbose: Optional[bool] = None,
        scrape_metrics: Optional[bool] = None,
        local_code_path: Optional[str] = None,
        auto_build: Optional[bool] = None,
        param_file_secrets: Optional[dict[str, str]] = None,
        notifications: Optional[list[mlrun.model.Notification]] = None,
        returns: Optional[list[Union[str, dict[str, str]]]] = None,
        state_thresholds: Optional[dict[str, int]] = None,
        reset_on_run: Optional[bool] = None,
        retry: Optional[Union[mlrun.model.Retry, dict]] = None,
    ) -> "mlrun.run.RunObject":
        # do not allow local function to be scheduled
        if schedule is not None:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Unexpected {schedule=} parameter for local function execution"
            )

        self.enrich_runtime(runtime, project)
        run = self._create_run_object(task)

        if self._is_run_local:
            runtime = self._create_local_function_for_execution(
                runtime=runtime,
                run=run,
                local_code_path=local_code_path,
                project=project,
                name=name,
                workdir=workdir,
                handler=handler,
                reset_on_run=reset_on_run,
            )

        # sanity check
        elif runtime._is_remote:
            message = "Remote function cannot be executed locally"
            logger.error(
                message,
                is_remote=runtime._is_remote,
                local=self._is_run_local,
                runtime=runtime.to_dict(),
            )
            raise mlrun.errors.MLRunRuntimeError(message)

        run = self._enrich_run(
            runtime=runtime,
            run=run,
            handler=handler,
            project_name=project,
            name=name,
            params=params,
            inputs=inputs,
            returns=returns,
            hyperparams=hyperparams,
            hyper_param_options=hyper_param_options,
            verbose=verbose,
            scrape_metrics=scrape_metrics,
            output_path=output_path,
            workdir=workdir,
            notifications=notifications,
            state_thresholds=state_thresholds,
            retry=retry,
        )
        self._validate_run(runtime, run)
        result = self._execute(
            runtime=runtime,
            run=run,
        )

        return result

    def _execute(
        self,
        runtime: "mlrun.runtimes.BaseRuntime",
        run: Optional[Union["mlrun.run.RunTemplate", "mlrun.run.RunObject"]] = None,
    ):
        if (
            "V3IO_USERNAME" in os.environ
            and mlrun_constants.MLRunInternalLabels.v3io_user not in run.metadata.labels
        ):
            run.metadata.labels[mlrun_constants.MLRunInternalLabels.v3io_user] = (
                os.environ.get("V3IO_USERNAME")
            )

        # store function object in db unless running from within a run pod
        if not runtime.is_child:
            logger.info(
                "Storing function",
                name=run.metadata.name,
                uid=run.metadata.uid,
                db=runtime.spec.rundb,
            )
            self._store_function(runtime, run)

        execution = mlrun.run.MLClientCtx.from_dict(
            run.to_dict(),
            runtime._get_db(),
            autocommit=False,
            is_api=False,
            store_run=False,
        )

        # create task generator (for child runs) from spec
        task_generator = mlrun.runtimes.generators.get_generator(run.spec, execution)
        if task_generator:
            # verify valid task parameters
            tasks = task_generator.generate(run)
            for task in tasks:
                self._validate_run_params(task.spec.parameters)

        # post verifications, store execution in db and run pre run hooks
        execution.store_run()
        runtime._pre_run(run, execution)  # hook for runtime specific prep

        last_err = None
        # If the runtime is nested, it means the hyper-run will run within a single instance of the run.
        # So while in the API, we consider the hyper-run as a single run, and then in the runtime itself when the
        # runtime is now a local runtime and therefore `self._is_nested == False`, we run each task as a separate run by
        # using the task generator
        # TODO client-server separation might not need the not runtime._is_nested anymore as this executed local func
        if task_generator and not runtime._is_nested:
            # multiple runs (based on hyper params or params file)
            runner = runtime._run_many
            if hasattr(runtime, "_parallel_run_many") and task_generator.use_parallel():
                runner = runtime._parallel_run_many
            results = runner(task_generator, execution, run)
            mlrun.runtimes.utils.results_to_iter(results, run, execution)
            result = execution.to_dict()
            result = runtime._update_run_state(result, task=run)

        else:
            # single run
            try:
                resp = runtime._run(run, execution)
                result = runtime._update_run_state(resp, task=run)
            except mlrun.runtimes.base.RunError as err:
                last_err = err
                result = runtime._update_run_state(task=run, err=err)

        self._push_notifications(run, runtime)

        # run post run hooks
        runtime._post_run(result, execution)  # hook for runtime specific cleanup

        return self._wrap_run_result(runtime, result, run, err=last_err)

    def _create_local_function_for_execution(
        self,
        runtime: "mlrun.runtimes.BaseRuntime",
        run: "mlrun.run.RunObject",
        local_code_path: Optional[str] = None,
        project: Optional[str] = "",
        name: Optional[str] = "",
        workdir: Optional[str] = "",
        handler: Optional[str] = None,
        reset_on_run: Optional[bool] = None,
    ):
        project = project or runtime.metadata.project
        function_name = name or runtime.metadata.name
        command, args = self._resolve_local_code_path(local_code_path)
        if command:
            function_name = name or pathlib.Path(command).stem

        meta = mlrun.model.BaseMetadata(function_name, project=project)

        command, loaded_runtime = mlrun.run.load_func_code(
            command or runtime, workdir, name=name
        )
        # loaded_runtime is loaded from runtime or yaml file, if passed a command it should be None,
        # so we keep the current runtime for enrichment
        runtime = loaded_runtime or runtime
        if loaded_runtime:
            if run:
                handler = handler or run.spec.handler
            handler = handler or runtime.spec.default_handler or ""
            meta = runtime.metadata.copy()
            meta.name = function_name or meta.name
            meta.project = project or meta.project

        # if the handler has module prefix force "local" (vs "handler") runtime
        kind = "local" if isinstance(handler, str) and "." in handler else ""

        # Create temporary local function for execution
        fn = mlrun.new_function(meta.name, command=command, args=args, kind=kind)
        fn.metadata = meta
        setattr(fn, "_is_run_local", True)
        if workdir:
            fn.spec.workdir = str(workdir)

        fn.spec.allow_empty_resources = runtime.spec.allow_empty_resources
        if runtime:
            # copy the code/base-spec to the local function (for the UI and code logging)
            fn.spec.description = runtime.spec.description
            fn.spec.build = runtime.spec.build
            serving_spec = getattr(runtime.spec, "serving_spec", None)
            if serving_spec:
                environ["SERVING_SPEC_ENV"] = serving_spec

        run.spec.handler = handler
        run.spec.reset_on_run = reset_on_run
        return fn

    @staticmethod
    def _resolve_local_code_path(local_code_path: str) -> (str, list[str]):
        command = None
        args = []
        if local_code_path:
            command = local_code_path
            if command:
                sp = command.split()
                # split command and args
                command = sp[0]
                if len(sp) > 1:
                    args = sp[1:]
        return command, args

    def _validate_run(
        self,
        runtime: "mlrun.runtimes.BaseRuntime",
        run: "mlrun.run.RunObject",
    ):
        super()._validate_run(runtime, run)
        if self._is_run_local and run.spec.retry.count:
            logger.warning(
                "Retry is not supported for local runs, ignoring retry settings",
                retry=run.spec.retry.to_dict(),
            )
            run.spec.retry.count = 0

    def _push_notifications(
        self, runobj: "mlrun.run.RunObject", runtime: "mlrun.runtimes.BaseRuntime"
    ):
        if not self._run_has_valid_notifications(runobj):
            return
        # TODO: add store_notifications API endpoint so we can store notifications pushed from the
        #       SDK for documentation purposes.
        # The run is local, so we can assume that watch=True, therefore this code runs
        # once the run is completed, and we can just push the notifications.
        # Only push from jupyter, not from the CLI.
        # "handler" and "dask" kinds are special cases of local runs which don't set local=True
        if self._is_run_local or runtime.kind in ["handler"]:
            mlrun.utils.notifications.NotificationPusher([runobj]).push()
        elif runtime.kind in ["dask"]:
            runtime._get_db().push_run_notifications(
                uid=runobj.metadata.uid, project=runobj.metadata.project
            )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Union

import numpy as np
import torch
from torch import Tensor
from torch.nn import Module

import mlrun
from mlrun.data_types import ValueType

from .._dl_common import DLTypes, DLUtils


class PyTorchTypes(DLTypes):
    """
    Typing hints for the PyTorch framework.
    """

    # Every model in PyTorch must inherit from torch.nn.Module:
    ModelType = Module

    # Supported types of loss and metrics values:
    MetricValueType = Union[int, float, np.ndarray, Tensor]

    # Supported types of metrics:
    MetricFunctionType = Union[Callable[[Tensor, Tensor], MetricValueType], Module]


class PyTorchUtils(DLUtils):
    """
    Utilities functions for the PyTorch framework.
    """

    @staticmethod
    def convert_value_type_to_torch_dtype(
        value_type: str,
    ) -> torch.dtype:
        """
        Get the 'torch.dtype' equivalent to the given MLRun data type.

        :param value_type: The MLRun value type to convert to torch data type.

        :return: The 'torch.dtype' equivalent to the given MLRun data type.

        :raise MLRunInvalidArgumentError: If torch is not supporting the given value type.
        """
        # Initialize the mlrun to torch data type conversion map:
        conversion_map = {
            ValueType.BOOL: torch.bool,
            ValueType.INT8: torch.int8,
            ValueType.INT16: torch.int16,
            ValueType.INT32: torch.int32,
            ValueType.INT64: torch.int64,
            ValueType.UINT8: torch.uint8,
            ValueType.BFLOAT16: torch.bfloat16,
            ValueType.FLOAT16: torch.float16,
            ValueType.FLOAT: torch.float32,
            ValueType.DOUBLE: torch.float64,
        }

        # Convert and return:
        if value_type in conversion_map:
            return conversion_map[value_type]
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"The ValueType given is not supported in torch: '{value_type}'."
        )

    @staticmethod
    def convert_torch_dtype_to_value_type(torch_dtype: Union[torch.dtype, str]) -> str:
        """
        Convert the given torch data type to MLRun value type. All the CUDA supported data types are supported. For
        more information regarding torch data types, visit: https://pytorch.org/docs/stable/tensors.html#data-types

        :param torch_dtype: The torch data type to convert to MLRun's value type. Expected to be a 'torch.dtype' or
                            'str'.

        :return: The MLRun value type converted from the given data type.

        :raise MLRunInvalidArgumentError: If the torch data type is not supported by MLRun.
        """
        # Initialize the torch to mlrun data type conversion map:
        conversion_map = {
            str(torch.bool): ValueType.BOOL,
            str(torch.int8): ValueType.INT8,
            str(torch.short): ValueType.INT16,
            str(torch.int16): ValueType.INT16,
            str(torch.int): ValueType.INT32,
            str(torch.int32): ValueType.INT32,
            str(torch.long): ValueType.INT64,
            str(torch.int64): ValueType.INT64,
            str(torch.uint8): ValueType.UINT8,
            str(torch.bfloat16): ValueType.BFLOAT16,
            str(torch.half): ValueType.FLOAT16,
            str(torch.float16): ValueType.FLOAT16,
            str(torch.float): ValueType.FLOAT,
            str(torch.float32): ValueType.FLOAT,
            str(torch.double): ValueType.DOUBLE,
            str(torch.float64): ValueType.DOUBLE,
        }

        # Parse the given torch data type to string:
        if isinstance(torch_dtype, torch.dtype):
            torch_dtype = str(torch_dtype)

        # Convert and return:
        if torch_dtype in conversion_map:
            return conversion_map[torch_dtype]
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"MLRun value type is not supporting the given torch data type: '{torch_dtype}'."
        )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .._common import LoggingMode, Producer
from .plan import MLPlanStages


class MLProducer(Producer):
    """
    Class for handling production of artifact plans during a run.
    """

    def is_probabilities_required(self) -> bool:
        """
        Check if probabilities are required in order to produce some of the artifacts.

        :return: True if probabilities are required by at least one plan and False otherwise.
        """
        return any(plan.need_probabilities for plan in self._plans)

    def produce_stage(
        self, stage: MLPlanStages, is_probabilities: bool = False, **kwargs
    ):
        """
        Produce the artifacts ready at the given stage and log them.

        :param stage:            The current stage to log at.
        :param is_probabilities: True if the 'y_pred' is a prediction of probabilities (from 'predict_proba') and False
                                 if not. Default: False.
        :param kwargs:           All of the required produce arguments to pass onto the plans.
        """
        # Produce all the artifacts according to the given stage:
        self._produce_artifacts(
            stage=stage, is_probabilities=is_probabilities, **kwargs
        )

        # Log if a context is available:
        if self._context is not None:
            # Log the artifacts in queue:
            self._log_artifacts()
            # Commit:
            self._context.commit(completed=False)

    def _produce_artifacts(
        self, stage: MLPlanStages, is_probabilities: bool = False, **kwargs
    ):
        """
        Go through the plans and check if they are ready to be produced in the given stage of the run. If they are,
        the logger will pass all the arguments to the 'plan.produce' method and collect the returned artifact.

        :param stage:            The stage to produce the artifact to check if its ready.
        :param is_probabilities: True if the 'y_pred' that will be sent to 'produce' is a prediction of probabilities
                                 (from 'predict_proba') and False if not. Default: False.
        :param kwargs:           All of the required produce arguments to pass onto the plans.
        """
        # Initialize a new list of plans for all the plans that will still need to be produced:
        plans = []

        # Go ver the plans to produce their artifacts:
        for plan in self._plans:
            # Check if the plan is ready:
            if plan.is_ready(stage=stage, is_probabilities=is_probabilities):
                # Produce the artifact:
                self._not_logged_artifacts = {
                    **self._not_logged_artifacts,
                    **plan.produce(**kwargs),
                }
                # If the plan should not be produced again, continue to the next one so it won't be collected:
                if not plan.is_reproducible():
                    continue
            # Collect the plan to produce it later (or again if reproducible):
            plans.append(plan)

        # Clear the old plans:
        self._plans = plans

        # Add evaluation prefix if in Evaluation mode:
        if self._mode == LoggingMode.EVALUATION:
            self._not_logged_artifacts = {
                f"evaluation-{key}": value
                for key, value in self._not_logged_artifacts.items()
            }
            for artifact in self._not_logged_artifacts.values():
                artifact.key = f"evaluation-{artifact.key}"



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "ingest",
    "delete_feature_set",
    "delete_feature_vector",
    "get_feature_set",
    "get_feature_vector",
    "Feature",
    "Entity",
    "FeatureSet",
    "FeatureVector",
    "RunConfig",
    "OfflineVectorResponse",
    "OnlineVectorService",
    "FixedWindowType",
]


from ..data_types import InferOptions, ValueType
from ..features import Entity, Feature
from .api import (
    delete_feature_set,
    delete_feature_vector,
    get_feature_set,
    get_feature_vector,
    ingest,
)
from .common import RunConfig
from .feature_set import FeatureSet
from .feature_vector import (
    FeatureVector,
    FixedWindowType,
)
from .feature_vector_utils import (
    JoinGraph,
    OfflineVectorResponse,
    OnlineVectorService,
)
